from itertools import zip_longest

from numpy import arange, isin, sort, concatenate

from hypothesis import given, assume, example
from hypothesis.strategies import one_of

from pytest import raises

from ..chunking import ChunkSize
from ..tuple import Tuple
from ..ndindex import ndindex

from .helpers import assert_equal, chunk_sizes, chunk_shapes, prod, ints, slices, ndindices

def test_ChunkSize_constructor():
    raises(TypeError, lambda: ChunkSize(Tuple(1, 2, 3)))
    raises(TypeError, lambda: ChunkSize(1, 2, 3))
    raises(TypeError, lambda: ChunkSize(1))
    raises(ValueError, lambda: ChunkSize((-1, 2, 3)))
    raises(ValueError, lambda: ChunkSize((0, 2, 3)))
    raises(NotImplementedError, lambda: ChunkSize((None, 2, 3)))

@given(chunk_sizes())
def test_ChunkSize_eq(chunk_size_tuple):
    chunk_size = ChunkSize(chunk_size_tuple)
    new = type(chunk_size)(*chunk_size.args)

    assert chunk_size == chunk_size_tuple
    assert chunk_size_tuple == chunk_size
    assert new == chunk_size
    assert chunk_size == new
    assert new == chunk_size_tuple
    assert chunk_size_tuple == new

    assert hash(new) == hash(chunk_size)
    assert not (chunk_size == 'a')
    assert not ('a' == chunk_size)
    assert (chunk_size != 'a')
    assert ('a' != chunk_size)

    h = hash(chunk_size_tuple)
    assert hash(chunk_size) == h

@given(chunk_sizes(), one_of(ints(), slices()))
def test_ChunkSize_args(chunk_size_tuple, idx):
    chunk_size = ChunkSize(chunk_size_tuple)
    assert chunk_size.args == (chunk_size_tuple,)

    try:
        ndindex(idx)
    except ValueError:
        # Filter out invalid slices (TODO: do this in the strategy)
        assume(False)

    # Should index the same way
    # TODO: Refactor check_same() so we can use that
    try:
        chunk_size_idx = chunk_size[idx]
    except IndexError:
        try:
            tuple_idx = chunk_size_tuple[idx]
        except IndexError:
            pass
        else:
            raise AssertionError("ChunkSize raised but tuple did not")
    else:
        tuple_idx = chunk_size_tuple[idx]
        assert chunk_size_idx == tuple_idx

@given(chunk_sizes())
def test_ChunkSize_tuple(chunk_size_tuple):
    # Test that ChunkSize behaves like a tuple
    chunk_size = ChunkSize(chunk_size_tuple)
    assert tuple(chunk_size) == chunk_size_tuple

def test_indices_error():
    raises(ValueError, lambda: next(ChunkSize((1, 2)).indices((1, 2, 3))))

@given(chunk_sizes(), chunk_shapes)
def test_num_chunks(chunk_size, shape):
    chunk_size = ChunkSize(chunk_size)
    assert chunk_size.num_chunks(shape) == len(list(chunk_size.indices(shape)))

@given(chunk_sizes(), chunk_shapes)
def test_indices(chunk_size, shape):
    chunk_size = ChunkSize(chunk_size)
    indices = chunk_size.indices(shape)
    size = prod(shape)
    a = arange(size).reshape(shape)

    subarrays = []
    for idx in indices:
        # The indices should be fully expanded
        assert idx.expand(shape) == idx
        # Except for indices at the edges, they should index a full chunk
        if not any(s.stop == i for s, i in zip(idx.args, shape)):
            assert idx.newshape(shape) == chunk_size
        # Make sure they can be indexed
        subarrays.append(a[idx.raw])
    # Check that indices together index every element of the array exactly
    # once.
    elements = [i for x in subarrays for i in x.flatten()]
    assert sorted(elements) == list(range(size))

def test_as_subchunks_error():
    raises(ValueError, lambda: next(ChunkSize((1, 2)).as_subchunks(..., (1, 2, 3))))

@example((1,), (0,), ...)
@example((2, 2), (5, 5), (0, 3))
@example((2, 2), (5, 5), (slice(0, 5, 2), slice(0, 5, 3)))
@example((2, 2), (5, 5), ([0, 0],))
@given(chunk_sizes(), chunk_shapes, ndindices)
def test_as_subchunks(chunk_size, shape, idx):
    chunk_size = ChunkSize(chunk_size)
    size = prod(shape)
    a = arange(size).reshape(shape)
    idx = ndindex(idx)

    try:
        idx.reduce(shape)
    except IndexError:
        assume(False)

    full_idx = a[idx.raw]

    try:
        subarrays = []
        fast = chunk_size.as_subchunks(idx, shape)
        slow = chunk_size.as_subchunks(idx, shape, _force_slow=True)
        for c, cslow in zip_longest(fast, slow):
            assert c == cslow
            index = idx.as_subindex(c)
            chunk = a[c.raw]
            subchunk = chunk[index.raw]
            # Not empty
            assert subchunk.size > 0
            # Indexes the right elements (c.f. test_as_subindex)
            assert_equal(subchunk.flatten(), full_idx[isin(full_idx, chunk)])
            subarrays.append(subchunk)
    except NotImplementedError:
        return

    # Picks all elements
    if subarrays:
        elements = concatenate([x.flatten() for x in subarrays])
    else:
        elements = arange(0)
    assert_equal(sort(elements), sort(full_idx.flatten()))



from numpy import intp, array, int16

from pytest import raises

from ..array import ArrayIndex
from ..integerarray import IntegerArray

from .helpers import assert_equal

# Everything else is tested in the subclasses

def test_ArrayIndex():
    raises(TypeError, lambda: ArrayIndex([]))

def test_attributes():
    a = array([[0, 1], [1, 0]])

    idx = IntegerArray(a)
    assert_equal(idx.array, array(a, dtype=intp))
    assert idx.dtype == intp
    assert idx.ndim == a.ndim == 2
    assert idx.shape == a.shape == (2, 2)
    assert idx.size == a.size == 4

def test_cast_raises():
    with raises(TypeError):
        a = array([[0, 1], [1, 0]])
        idx = IntegerArray(a)
        assert array(idx) == idx

def test_copy():
    idx = IntegerArray([1, 2])
    idx2 = IntegerArray(idx.raw)
    assert idx.raw is not idx2.raw
    idx3 = IntegerArray(idx.raw, _copy=False)
    assert idx.raw is idx3.raw
    raises(ValueError, lambda: IntegerArray([], _copy=False))
    raises(ValueError, lambda: IntegerArray(array([1], dtype=int16), _copy=False))



import sys
from itertools import chain
from functools import reduce
from operator import mul

from numpy import intp, bool_, array
import numpy.testing

from pytest import fail

from hypothesis.strategies import (integers, none, one_of, lists, just,
                                   builds, shared, composite)
from hypothesis.extra.numpy import arrays

from ..ndindex import ndindex

# Hypothesis strategies for generating indices. Note that some of these
# strategies are nominally already defined in hypothesis, but we redefine them
# here because the hypothesis definitions are too restrictive. For example,
# hypothesis's slices strategy does not generate slices with negative indices.
# Similarly, hypothesis.extra.numpy.basic_indices only generates tuples.

# np.prod has overflow and math.prod is Python 3.8+ only
def prod(seq):
    return reduce(mul, seq, 1)

nonnegative_ints = integers(0, 10)
negative_ints = integers(-10, -1)
ints = lambda: one_of(nonnegative_ints, negative_ints)

def slices(start=one_of(none(), ints()), stop=one_of(none(), ints()),
           step=one_of(none(), ints())):
    return builds(slice, start, stop, step)

ellipses = lambda: just(...)
newaxes = lambda: just(None)

# hypotheses.strategies.tuples only generates tuples of a fixed size
def tuples(elements, *, min_size=0, max_size=None, unique_by=None, unique=False):
    return lists(elements, min_size=min_size, max_size=max_size,
                 unique_by=unique_by, unique=unique).map(tuple)

shapes = tuples(integers(0, 10)).filter(
             # numpy gives errors with empty arrays with large shapes.
             # See https://github.com/numpy/numpy/issues/15753
             lambda shape: prod([i for i in shape if i]) < 100000)

_short_shapes = tuples(integers(0, 10)).filter(
             # numpy gives errors with empty arrays with large shapes.
             # See https://github.com/numpy/numpy/issues/15753
             lambda shape: prod([i for i in shape if i]) < 1000)

# We need to make sure shapes for boolean arrays are generated in a way that
# makes them related to the test array shape. Otherwise, it will be very
# difficult for the boolean array index to match along the test array, which
# means we won't test any behavior other than IndexError.

# short_shapes should be used in place of shapes in any test function that
# uses ndindices, boolean_arrays, or tuples
short_shapes = shared(_short_shapes)

_integer_arrays = arrays(intp, short_shapes)
integer_scalars = arrays(intp, ()).map(lambda x: x[()])
integer_arrays = one_of(integer_scalars, _integer_arrays.flatmap(lambda x: one_of(just(x), just(x.tolist()))))

@composite
def subsequences(draw, sequence):
    seq = draw(sequence)
    start = draw(integers(0, max(0, len(seq)-1)))
    stop = draw(integers(start, len(seq)))
    return seq[start:stop]

_boolean_arrays = arrays(bool_, one_of(subsequences(short_shapes), short_shapes))
boolean_scalars = arrays(bool_, ()).map(lambda x: x[()])
boolean_arrays = one_of(boolean_scalars, _boolean_arrays.flatmap(lambda x: one_of(just(x), just(x.tolist()))))

def _doesnt_raise(idx):
    try:
        ndindex(idx)
    except (IndexError, ValueError, NotImplementedError):
        return False
    return True

Tuples = tuples(one_of(ellipses(), ints(), slices(), newaxes(),
                       integer_arrays, boolean_arrays)).filter(_doesnt_raise)

ndindices = one_of(
    ints(),
    slices(),
    ellipses(),
    newaxes(),
    Tuples,
    integer_arrays,
    boolean_arrays,
).filter(_doesnt_raise)

def assert_equal(actual, desired, err_msg='', verbose=True):
    """
    Same as numpy.testing.assert_equal except it also requires the shapes and
    dtypes to be equal.

    """
    numpy.testing.assert_equal(actual, desired, err_msg=err_msg,
                               verbose=verbose)
    assert actual.shape == desired.shape, err_msg or f"{actual.shape} != {desired.shape}"
    assert actual.dtype == desired.dtype, err_msg or f"{actual.dtype} != {desired.dtype}"

def check_same(a, idx, raw_func=lambda a, idx: a[idx],
               ndindex_func=lambda a, index: a[index.raw],
               same_exception=True, assert_equal=assert_equal):
    """
    Check that a raw index idx produces the same result on an array a before
    and after being transformed by ndindex.

    Tests that raw_func(a, idx) == ndindex_func(a, ndindex(idx)) or that they
    raise the same exception. If same_exception=False, it will still check
    that they both raise an exception, but will not require the exception type
    and message to be the same.

    By default, raw_func(a, idx) is a[idx] and ndindex_func(a, index) is
    a[index.raw].

    The assert_equal argument changes the function used to test equality. By
    default it is the custom assert_equal() function in this file that extends
    numpy.testing.assert_equal. If the func functions return something other
    than arrays, assert_equal should be set to something else, like

        def assert_equal(x, y):
            assert x == y

    """
    exception = None
    try:
        # Handle list indices that NumPy treats as tuple indices with a
        # deprecation warning. We want to test against the post-deprecation
        # behavior.
        e_inner = None
        try:
            try:
                a_raw = raw_func(a, idx)
            except Warning as w:
                if ("Using a non-tuple sequence for multidimensional indexing is deprecated" in w.args[0]):
                    idx = array(idx)
                    a_raw = raw_func(a, idx)
                elif "Out of bound index found. This was previously ignored when the indexing result contained no elements. In the future the index error will be raised. This error occurs either due to an empty slice, or if an array has zero elements even before indexing." in w.args[0]:
                    same_exception = False
                    raise IndexError
                else: # pragma: no cover
                    fail(f"Unexpected warning raised: {w}")
        except Exception:
            _, e_inner, _ = sys.exc_info()
        if e_inner:
            raise e_inner
    except Exception as e:
        exception = e

    try:
        index = ndindex(idx)
        a_ndindex = ndindex_func(a, index)
    except Exception as e:
        if not exception:
            fail(f"Raw form does not raise but ndindex form does ({e!r}): {index})") # pragma: no cover
        if same_exception:
            assert type(e) == type(exception), (e, exception)
            assert e.args == exception.args, (e.args, exception.args)
    else:
        if exception:
            fail(f"ndindex form did not raise but raw form does ({exception!r}): {index})") # pragma: no cover

    if not exception:
        assert_equal(a_raw, a_ndindex)


def iterslice(start_range=(-10, 10),
               stop_range=(-10, 10),
               step_range=(-10, 10),
               one_two_args=True
):
    # one_two_args is unnecessary if the args are being passed to slice(),
    # since slice() already canonicalizes missing arguments to None. We do it
    # for Slice to test that behavior.
    if one_two_args:
        for start in chain(range(*start_range), [None]):
            yield (start,)

        for start in chain(range(*start_range), [None]):
            for stop in chain(range(*stop_range), [None]):
                yield (start, stop)

    for start in chain(range(*start_range), [None]):
        for stop in chain(range(*stop_range), [None]):
            for step in chain(range(*step_range), [None]):
                yield (start, stop, step)


chunk_shapes = shared(shapes)

@composite
def chunk_sizes(draw, shapes=chunk_shapes):
    shape = draw(shapes)
    return draw(tuples(integers(1, 10), min_size=len(shape),
                       max_size=len(shape)).filter(lambda shape: prod(shape) < 10000))



import warnings

from numpy import ndarray, asarray, integer, bool_, array2string, empty, intp

from .ndindex import NDIndex, asshape

class ArrayIndex(NDIndex):
    """
    Superclass for array indices

    This class should not be instantiated directly. Rather, use one of its
    subclasses, :class:`~.IntegerArray` or :class:`~.BooleanArray`.

    To subclass this, define the `dtype` attribute, as well as all the usual
    ndindex methods.
    """
    # Subclasses should redefine this
    dtype = None

    def _typecheck(self, idx, shape=None, _copy=True):
        if self.dtype is None:
            raise TypeError("Do not instantiate the superclass ArrayIndex directly")

        if shape is not None:
            if idx != []:
                raise ValueError("The shape argument is only allowed for empty arrays (idx=[])")
            shape = asshape(shape)
            if 0 not in shape:
                raise ValueError("The shape argument must be an empty shape")
            idx = empty(shape, dtype=self.dtype)

        if isinstance(idx, (list, ndarray, bool, integer, int, bool_)):
            # Ignore deprecation warnings for things like [1, []]. These will be
            # filtered out anyway since they produce object arrays.
            with warnings.catch_warnings(record=True):
                a = asarray(idx)
                if a is idx and _copy:
                    a = a.copy()
                if isinstance(idx, list) and 0 in a.shape:
                    if not _copy:
                        raise ValueError("_copy=False is not allowed with list input")
                    a = a.astype(self.dtype)
            if self.dtype == intp and issubclass(a.dtype.type, integer):
                if a.dtype != self.dtype:
                    if not _copy:
                        raise ValueError("If _copy=False, the input array dtype must already be intp")
                    a = a.astype(self.dtype)
            if a.dtype != self.dtype:
                raise TypeError(f"The input array to {self.__class__.__name__} must have dtype {self.dtype.__name__}, not {a.dtype}")
            a.flags.writeable = False
            return (a,)
        raise TypeError(f"{self.__class__.__name__} must be created with an array with dtype {self.dtype.__name__}")

    # These will allow array == ArrayIndex to give True or False instead of
    # returning an array.
    __array_ufunc__ = None
    def __array_function__(self, func, types, args, kwargs):
        return NotImplemented

    def __array__(self):
        raise TypeError(f"Cannot convert {self.__class__.__name__} to an array. Use .array instead.")


    @property
    def raw(self):
        return self.args[0]

    @property
    def array(self):
        """
        Return the NumPy array of self.

        This is the same as `self.args[0]`.

        >>> from ndindex import IntegerArray, BooleanArray
        >>> IntegerArray([0, 1]).array
        array([0, 1])
        >>> BooleanArray([False, True]).array
        array([False, True])

        """
        return self.args[0]

    @property
    def shape(self):
        """
        Return the shape of the array of self.

        This is the same as `self.array.shape`. Note that this is **not** the
        same as the shape of an array that is indexed by `self`. Use
        :meth:`~.NDIndex.newshape` to get that.

        >>> from ndindex import IntegerArray, BooleanArray
        >>> IntegerArray([[0], [1]]).shape
        (2, 1)
        >>> BooleanArray([[False], [True]]).shape
        (2, 1)

        """
        return self.array.shape

    @property
    def ndim(self):
        """
        Return the number of dimensions of the array of self.

        This is the same as `self.array.ndim`. Note that this is **not** the
        same as the number of dimensions of an array that is indexed by
        `self`. Use `len` on :meth:`~.NDIndex.newshape` to get that.

        >>> from ndindex import IntegerArray, BooleanArray
        >>> IntegerArray([[0], [1]]).ndim
        2
        >>> BooleanArray([[False], [True]]).ndim
        2

        """
        return self.array.ndim

    @property
    def size(self):
        """
        Return the number of elements of the array of self.

        This is the same as `self.array.size`. Note that this is **not** the
        same as the number of elements of an array that is indexed by `self`.
        Use `np.prod` on :meth:`~.NDIndex.newshape` to get that.

        >>> from ndindex import IntegerArray, BooleanArray
        >>> IntegerArray([[0], [1]]).size
        2
        >>> BooleanArray([[False], [True]]).size
        2

        """
        return self.array.size

    # The repr form recreates the object. The str form gives the truncated
    # array string and is explicitly non-valid Python (doesn't have commas).
    def __repr__(self):
        if 0 not in self.shape:
            arg = repr(self.array.tolist())
        else:
            arg = f"[], shape={self.shape}"
        return f"{self.__class__.__name__}({arg})"

    def __str__(self):
        return (self.__class__.__name__
                + "("
                + array2string(self.array).replace('\n', '')
                + ")")

    def __hash__(self):
        return hash(self.array.tobytes())



from numpy import bool_, count_nonzero, ndarray

from .array import ArrayIndex
from .ndindex import asshape

class BooleanArray(ArrayIndex):
    """
    Represents a boolean array index (also known as a mask).

    If `idx` is an n-dimensional boolean array with shape `s = (s1, ..., sn)`
    and `a` is an array of shape `s = (s1, ..., sn, ..., sm)`, `a[idx]`
    replaces the first `n` dimensions of `a` with a single dimensions of size
    `np.nonzero(idx)`, where each entry is included if the corresponding
    element of `idx` is True. The axes in the index shape should match the
    corresponding axes in the array shape or be 0, or the index produces
    IndexError.

    The typical way of creating a mask is to use boolean operations on an
    array, then index the array with that. For example, if `a` is an array of
    integers, `a[a > 0]` will produces a flat array of the elements of `a`
    that are positive.

    Some important things to note about boolean array index semantics:

    1. A boolean array index will remove as many dimensions as the index has,
       and replace them with a single flat dimension which is the size of the
       number of `True` elements in the index.

    2. A boolean array index `idx` works the same as the integer array index
       `np.nonzero(idx)`. In particular, the elements of the index are always
       iterated in row-major, C-style order. This does not apply to
       0-dimensional boolean indices.

    3. A 0-dimensional boolean index (i.e., just the scalar `True` or `False`)
       can still be thought of as removing 0 dimensions and adding a single
       dimension of length 1 for True or 0 for False. Hence, if `a` has shape
       `(s1, ..., sn)`, then `a[True]` has shape `(1, s1, ..., sn)`, and
       `a[False]` has shape `(0, s1, ..., sn)`.

    4. If a tuple index has multiple boolean arrays, they are broadcast
       together and iterated as a single array, similar to
       :class:`IntegerArray`. If a boolean array index `idx` is mixed with an
       integer array index in a tuple index, it is treated like
       `np.nonzero(idx)`.

    A list of booleans may also be used in place of a boolean array. Note
    that NumPy treats a direct list of integers as a tuple index, but this
    behavior is deprecated and will be replaced with integer array indexing in
    the future. ndindex always treats lists as arrays.

    >>> from ndindex import BooleanArray
    >>> import numpy as np
    >>> idx = BooleanArray([[ True,  True],
    ...                     [ True, False],
    ...                     [False, False],
    ...                     [False,  True],
    ...                     [False, False]])
    >>> a = np.arange(10).reshape((5, 2))
    >>> a[idx.raw]
    array([0, 1, 2, 7])

    .. note::

       `BooleanArray` does *not* represent an array, but rather an *array
       index*. It does not have most methods that `numpy.ndarray` has, and
       should not be used in array contexts. See the document on
       :any:`type-confusion` for more details.

    """

    dtype = bool_
    """
    The dtype of `BooleanArray` is `np.bool_`.
    """

    def __hash__(self):
        # Match the hash for scalar booleans. Otherwise, hash(True) won't
        # equal hash(ndindex(True)).
        if self.shape == ():
            return hash(self.array.any())
        return super().__hash__()

    @property
    def count_nonzero(self):
        """
        Returns the number of elements indexed by self.

        In general, if shapes match, when indexed by `self`, the first *n*
        dimensions of an array are replaced with a single dimension of size
        `count_nonzero`, where *n* is `self.shape`.

        This is the same as `np.count_nonzero(self.array)`. Note, to get the
        shape of an array indexed by self, use :meth:`newshape`, not this
        method.

        >>> from ndindex import BooleanArray
        >>> BooleanArray([True, False, True]).count_nonzero
        2
        """
        return count_nonzero(self.array)

    def reduce(self, shape=None, axis=0):
        """
        Reduce a `BooleanArray` index on an array of shape `shape`.

        The result will either be `IndexError` if the index is invalid for the
        given shape, or a `BooleanArray` index. Presently, no simplifications
        are done for BooleanArray: if `reduce()` does not produce an
        `IndexArray` the index returned will be the same as `self`.

        >>> from ndindex import BooleanArray
        >>> idx = BooleanArray([True, False])
        >>> idx.reduce((3,))
        Traceback (most recent call last):
        ...
        IndexError: boolean index did not match indexed array along dimension 0; dimension is 3 but corresponding boolean dimension is 2
        >>> idx.reduce((2,))
        BooleanArray([True, False])

        See Also
        ========

        .NDIndex.reduce
        .Tuple.reduce
        .Slice.reduce
        .ellipsis.reduce
        .Newaxis.reduce
        .Integer.reduce
        .IntegerArray.reduce

        """
        if shape is None:
            return self

        shape = asshape(shape)

        if len(shape) < self.ndim + axis:
            raise IndexError(f"too many indices for array: array is {len(shape)}-dimensional, but {self.ndim + axis} were indexed")

        for i in range(axis, axis+self.ndim):
            if self.shape[i-axis] != 0 and shape[i] != self.shape[i-axis]:

                raise IndexError(f"boolean index did not match indexed array along dimension {i}; dimension is {shape[i]} but corresponding boolean dimension is {self.shape[i-axis]}")

        return self

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        shape = asshape(shape)

        # reduce will raise IndexError if it should be raised
        self.reduce(shape)
        return (self.count_nonzero,) + shape[self.ndim:]

    def isempty(self, shape=None):
        if shape is not None:
            return 0 in self.newshape(shape)

        return self.count_nonzero == 0

    def as_subindex(self, index):
        from .tuple import Tuple

        if self in [True, False]:
            raise NotImplementedError("as_subindex is not supported for scalar boolean indices")
        return Tuple(*self.array.nonzero()).as_subindex(index)

    def broadcast_arrays(self):
        from .tuple import Tuple
        return Tuple(self).broadcast_arrays()

    def __eq__(self, other):
        if isinstance(other, (bool, bool_)):
            return self.shape == () and self.array == other
        if isinstance(other, BooleanArray):
            b = other.array
        elif isinstance(other, ndarray):
            b = other
        elif isinstance(other, list):
            try:
                b = BooleanArray(other)
            except TypeError:
                return False
        else:
            return False
        a = self.array
        return a.shape == b.shape and (a == b).all()

def _is_boolean_scalar(idx):
    """
    Determine if idx is a scalar boolean index.

    This is for internal usage only. Assumes idx is already an ndindex type.
    This is more performant than `idx in [True, False]`.
    """
    # TODO: Instead of this function, make BooleanScalar a separate class.
    return isinstance(idx, BooleanArray) and idx.shape == ()



from numpy import arange, prod, array

from hypothesis import given, example
from hypothesis.strategies import integers, one_of

from ..ndindex import ndindex
from ..array import ArrayIndex
from ..booleanarray import BooleanArray
from ..integerarray import IntegerArray
from ..integer import Integer
from ..tuple import Tuple
from .helpers import ndindices, check_same, short_shapes

@example((..., False, False), 1)
@example((True, False), 1)
@example((True, True), 1)
@example(array([False]), 0)
@example((slice(None, -1, None), array([[1]])), (1, 1))
@example((1, False), 0)
@example(True, 1)
@example(False, 1)
@example([[True, False], [False, False]], (2, 2, 3))
@given(ndindices, one_of(short_shapes, integers(0, 10)))
def test_broadcast_arrays_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = ndindex(idx)

    # It is possible for the original index to give an IndexError, but for
    # broadcast_arrays() may return an index that doesn't (but the other way
    # around should not happen).
    check = True
    try:
        a[index.raw]
    except IndexError as e:
        if "boolean index did not match indexed array" in e.args[0]:
            check = False
        if index.isempty() and "out of bounds" in e.args[0]:
            # Integers are bounds checked even when the resulting index
            # broadcasts to empty (but not so for IntegerArray).
            check = False
    except DeprecationWarning as w:
        if "Out of bound index found. This was previously ignored when the indexing result contained no elements. In the future the index error will be raised. This error occurs either due to an empty slice, or if an array has zero elements even before indexing." in w.args[0]:
            pass
        else: # pragma: no cover
            raise
    if check:
        check_same(a, index.raw, ndindex_func=lambda a, x:
                   a[x.broadcast_arrays().raw], same_exception=False)

    broadcasted = index.broadcast_arrays()

    if isinstance(index, (Tuple, BooleanArray)):
        assert isinstance(broadcasted, Tuple)
    else:
        assert broadcasted == index

    if isinstance(broadcasted, Tuple):
        arrays = [arg for arg in broadcasted.args if isinstance(arg,
                                                                ArrayIndex)
                  and arg not in [True, False]]
        if arrays:
            assert len(set([i.shape for i in arrays])) == 1
            assert not any(isinstance(i, Integer) for i in broadcasted.args)
            assert all(isinstance(i, IntegerArray) or i in [True, False] for
                       i in arrays)

        assert broadcasted.args.count(True) <= 1
        assert broadcasted.args.count(False) <= 1
        assert not (True in broadcasted.args and False in broadcasted.args)
        if True in broadcasted.args or False in broadcasted.args:
            assert index in [True, False] or True in index.args or False in index.args
    if index in [True, False]:
        assert broadcasted == Tuple(index)
    elif isinstance(index, BooleanArray):
        assert all(isinstance(i, IntegerArray) for i in broadcasted.args)



from pytest import raises

from numpy import arange, prod, array, full

from hypothesis import given, example
from hypothesis.strategies import integers, one_of

from ..ndindex import ndindex
from ..tuple import Tuple
from ..integer import Integer
from .helpers import ndindices, check_same, short_shapes

@example(..., 0)
@example((True,), ())
@example(([[True, False], [True, False]], [True, True], slice(0, 2)), ((2, 2, 2, 3, 3)))
@example((array([], dtype=bool),), (0, 0))
@example((False, False), ())
@example(array([], dtype=bool), 0)
@example((array([], dtype=bool),), 0)
@example(array([[[True], [False]]]), (1, 1, 2))
@example(full((1, 9), False), (3, 3))
@example(([0, 1], 0), (2, 2))
@example(([0, 0, 0], [0, 0]), (2, 2))
@example((0, None, 0, ..., 0, None, 0), (2, 2, 2, 2, 2, 2, 2))
@example((0, slice(None), ..., slice(None), 3), (2, 3, 4, 5, 6, 7))
@given(ndindices, one_of(short_shapes, integers(0, 10)))
def test_newshape_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    try:
        index = ndindex(idx)
    except IndexError:
        pass
    else:
        # Make sure ndindex input gives an error
        raises(TypeError, lambda: index.newshape(Tuple(2, 1)))
        raises(TypeError, lambda: index.newshape(Integer(2)))

    def raw_func(a, idx):
        return a[idx].shape

    def ndindex_func(a, index):
        return index.newshape(shape)

    def assert_equal(raw_shape, newshape):
        assert raw_shape == newshape


    check_same(a, idx, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



from numpy import arange, prod, array, intp, empty

from hypothesis import given, example
from hypothesis.strategies import integers, one_of

from ..ndindex import ndindex
from ..array import ArrayIndex
from ..booleanarray import BooleanArray
from ..integerarray import IntegerArray
from ..integer import Integer
from ..tuple import Tuple
from .helpers import ndindices, check_same, short_shapes

@example(True, (1,))
@example((Ellipsis, array([[ True,  True]])), (1, 2))
@example((..., 0, 0, False), 1)
@example((empty((0, 0), dtype=bool)), 0)
@example((0, empty((0, 0), dtype=bool)), 0)
@example((..., empty((0, 0), dtype=bool)), 0)
@example((..., 0, empty((0, 0), dtype=bool)), 0)
@example((array([], dtype=intp), 0), (0, 0))
@example((array([], dtype=intp), [0]), (0, 0))
@example((..., 0, array([], dtype=intp)), (0, 0))
@example((..., array(0), array([], dtype=intp)), (0, 0))
@example((False, False), ())
@example((-1, False), 1)
@example((..., False), ())
@example((array([0]),), ())
@example(([0, 1], 0), (2, 2))
@example((..., [0, 1], 0), (2, 2))
@example((..., None, 0), 1)
@example((0, 1, ..., 2, 3), (2, 3, 4, 5, 6, 7))
@example(None, 2)
@given(ndindices, one_of(short_shapes, integers(0, 10)))
def test_expand_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = ndindex(idx)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.expand(shape).raw],
               same_exception=False)

    try:
        expanded = index.expand(shape)
    except IndexError:
        pass
    else:
        assert isinstance(expanded, Tuple)
        assert ... not in expanded.args
        n_newaxis = 0
        boolean_scalars = 0
        if isinstance(idx, tuple):
            n_newaxis = index.args.count(None)
            if True in index.args or False in index.args:
                boolean_scalars = 1
        elif index == None:
            n_newaxis = 1
        elif index in [True, False]:
            boolean_scalars = 1
        if isinstance(shape, int):
            assert len(expanded.args) == 1 + n_newaxis + boolean_scalars
        else:
            assert len(expanded.args) == len(shape) + n_newaxis + boolean_scalars

        # Make sure arrays are broadcasted
        if any(isinstance(i, ArrayIndex) and i not in [True, False] for i in expanded.args):
            assert not any(isinstance(i, Integer) for i in expanded.args)
            assert not any(isinstance(i, BooleanArray) and i not in [True, False] for i in expanded.args)
            assert len({i.shape for i in expanded.args if isinstance(i,
                                                                     IntegerArray)}) in [0, 1]

        assert expanded.args.count(True) <= 1
        assert expanded.args.count(False) <= 1
        assert not (True in expanded.args and False in expanded.args)



from .ndindex import NDIndex, asshape, operator_index

class Integer(NDIndex):
    """
    Represents an integer index on an axis of an nd-array.

    Any object that implements `__index__` can be used as an integer index.

    >>> from ndindex import Integer
    >>> idx = Integer(1)
    >>> [0, 1, 2][idx.raw]
    1
    >>> idx = Integer(-3)
    >>> [0, 1, 2][idx.raw]
    0

    Note that `Integer` itself implements `__index__`, so it can be used as an
    index directly. However, it is still recommended to use `raw` for
    consistency, as this only works for `Integer`.

    .. note::

       `Integer` does *not* represent an integer, but rather an
       *integer index*. It does not have most methods that `int` has, and
       should not be used in non-indexing contexts. See the document on
       :any:`type-confusion` for more details.

    """
    def _typecheck(self, idx):
        idx = operator_index(idx)
        return (idx,)

    def __index__(self):
        return self.raw

    @property
    def raw(self):
        return self.args[0]

    def __len__(self):
        """
        Returns the number of elements indexed by `self`

        Since `self` is an integer index, this always returns 1. Note that
        integer indices always remove an axis.
        """
        return 1

    def reduce(self, shape=None, axis=0):
        """
        Reduce an Integer index on an array of shape `shape`.

        The result will either be `IndexError` if the index is invalid for the
        given shape, or an Integer index where the value is nonnegative.

        >>> from ndindex import Integer
        >>> idx = Integer(-5)
        >>> idx.reduce((3,))
        Traceback (most recent call last):
        ...
        IndexError: index -5 is out of bounds for axis 0 with size 3
        >>> idx.reduce((9,))
        Integer(4)

        See Also
        ========

        .NDIndex.reduce
        .Tuple.reduce
        .Slice.reduce
        .ellipsis.reduce
        .Newaxis.reduce
        .IntegerArray.reduce
        .BooleanArray.reduce

        """
        if shape is None:
            return self

        shape = asshape(shape, axis=axis)
        size = shape[axis]
        if self.raw >= size or -size > self.raw < 0:
            raise IndexError(f"index {self.raw} is out of bounds for axis {axis} with size {size}")

        if self.raw < 0:
            return self.__class__(size + self.raw)

        return self

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        shape = asshape(shape)

        # reduce will raise IndexError if it should be raised
        self.reduce(shape)
        return shape[1:]

    def as_subindex(self, index):
        from .ndindex import ndindex
        from .slice import Slice
        from .tuple import Tuple

        index = ndindex(index)

        if isinstance(index, Tuple):
            return Tuple(self).as_subindex(index)

        if not isinstance(index, Slice):
            raise NotImplementedError("Integer.as_subindex is only implemented for slices")

        if self == -1:
            s = Slice(self.args[0], None).as_subindex(index)
        else:
            s = Slice(self.args[0], self.args[0] + 1).as_subindex(index)
        if s == Slice(0, 0, 1):
            # The intersection is empty. There is no valid index we can return
            # here. We want an index that produces an empty array, but the
            # shape should be one less, to match a[self]. Since a[index] has
            # as many dimensions as a, there is no way to index a[index] so
            # that it gives one fewer dimension but is also empty. The best we
            # could do is to return a boolean array index array([False]),
            # which would replace the first dimension with a length 0
            # dimension. But
            #
            # 1. this isn't implemented yet,
            # 2. there are complications if this happens in multiple
            #    dimensions (it might not be possible to represent, I'm not
            #    sure), and
            # 3. Slice.as_subindex(Integer) also raises this exception in the
            #    case of an empty intersection (see the comment in that code).
            raise ValueError(f"{self} and {index} do not intersect")
        assert len(s) == 1
        return Integer(s.args[0])

    def isempty(self, shape=None):
        if shape is not None:
            return 0 in self.newshape(shape)

        return False

    def __eq__(self, other):
        if isinstance(other, Integer):
            return self.args == other.args
        try:
            other = operator_index(other)
        except TypeError:
            return False
        return self.args[0] == other

    def __hash__(self):
        return super().__hash__()



from numpy import prod, arange, array, bool_, empty, full

from hypothesis import given, example
from hypothesis.strategies import one_of, integers

from pytest import raises

from .helpers import boolean_arrays, short_shapes, check_same, assert_equal

from ..booleanarray import BooleanArray

def test_booleanarray_constructor():
    raises(ValueError, lambda: BooleanArray([False], shape=(1,)))
    raises(ValueError, lambda: BooleanArray([], shape=(1,)))
    raises(TypeError, lambda: BooleanArray([0]))
    raises(TypeError, lambda: BooleanArray(array(0.0)))
    raises(TypeError, lambda: BooleanArray((True,)))
    idx = BooleanArray(array([True], dtype=bool_))
    assert_equal(idx.array, array([True], dtype=bool_))

    idx = BooleanArray([], shape=(0, 1))
    assert_equal(idx.array, empty((0, 1), dtype=bool_))

    # Make sure the underlying array is immutable
    idx = BooleanArray([True])
    with raises(ValueError):
        idx.array[0] = False
    assert_equal(idx.array, array([True], dtype=bool_))

    # Make sure the underlying array is copied
    a = array([True, False])
    idx = BooleanArray(a)
    a[0] = False
    assert idx == BooleanArray([True, False])

@given(boolean_arrays, short_shapes)
def test_booleanarray_hypothesis(idx, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, idx)

@given(boolean_arrays, one_of(short_shapes, integers(0, 10)))
def test_booleanarray_reduce_no_shape_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = BooleanArray(idx)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.reduce().raw])

@example(full((1, 9), True), (3, 3))
@example(full((1, 9), False), (3, 3))
@given(boolean_arrays, one_of(short_shapes, integers(0, 10)))
def test_booleanarray_reduce_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = BooleanArray(idx)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.reduce(shape).raw])

    try:
        reduced = index.reduce(shape)
    except IndexError:
        pass
    else:
        # At present, reduce() always returns the same index if it doesn't
        # give an IndexError
        assert reduced == index

@given(boolean_arrays, one_of(short_shapes, integers(0, 10)))
def test_booleanarray_isempty_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = BooleanArray(idx)

    def raw_func(a, idx):
        return a[idx].size == 0

    def ndindex_func(a, index):
        return index.isempty(), index.isempty(shape)

    def assert_equal(raw_empty, ndindex_empty):
        isempty, isempty_shape = ndindex_empty

        # If isempty is True then a[t] should be empty
        if isempty:
            assert raw_empty, (index, shape)
        # We cannot test the converse with hypothesis. isempty may be False
        # but a[idx] could still be empty for this specific a (e.g., if a is
        # already itself empty).

        # If isempty is true with no shape it should be true for a specific
        # shape. The converse is not true because the indexed array could be
        # empty.
        if isempty:
            assert isempty_shape, (index, shape)

        # isempty() should always give the correct result for a specific
        # array after reduction
        assert isempty_shape == raw_empty, (index, shape)

    check_same(a, idx, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



"""
Tests are extremely important for ndindex. All operations should produce
correct results. We test this by checking against numpy arange (the array
values do not matter, so long as they are distinct).

There are three primary types of tests that we employ to verify this:

- Exhaustive tests. These test every possible value in some range. See for
  example test_slice. This is the best type of test, but unfortunately, it is
  often impossible to do due to combinatorial explosion.

- Hypothesis tests. Hypothesis is a library that can intelligently check a
  combinatorial search space. This requires writing hypothesis strategies that
  can generate all the relevant types of indices (see helpers.py). For more
  information on hypothesis, see
  https://hypothesis.readthedocs.io/en/latest/index.html.

- Explicit tests. These are hand crafted tests that test that the output of a
  function is some exact value. These are used when a property-based test is
  difficult to write, and it is simpler to test the exact output value. These
  tests still include a correctness check (check_same()) to make sure the
  hard-coded output is actually correct.

The basic idea in both cases is the same. Take the pure index and the
ndindex(index).raw, or in the case of a transformation, the before and after
raw index, and index an arange with them. If they do not give the same output
array, or do not both produce the same error, the code is not correct.

Why bother with hypothesis if the same thing is already tested exhaustively?
The main reason is that hypothesis is much better at producing human-readable
failure examples. When an exhaustive test fails, the failure will always be
from the first set of inputs in the loop that produces a failure. Hypothesis
on the other hand attempts to "shrink" the failure input to smallest input
that still fails. For example, a failing exhaustive slice test might give
Slice(-10, -9, -10) as a the failing example, but hypothesis would shrink it
to Slice(-2, -1, -1). Another reason for the duplication is that hypothesis
can sometimes test a slightly expanded test space without any additional
consequences. For example, test_slice_reduce_hypothesis() tests all types of
array shapes, whereas test_slice_reduce_exhaustive() tests only 1-dimensional
shapes. This doesn't affect things because hypothesis will always shrink large
shapes to a 1-dimensional shape in the case of a failure. Consequently every
exhaustive test should have a corresponding hypothesis test.

For things that can only be tested with hypothesis, you can use @example, to
force certain combinations to be tested. This is useful because we require
100% test coverage, and hypothesis's randomness can cause this to be flaky
otherwise.

"""

# Variable naming conventions in the tests:

# a: numpy arange. May be reshaped to be multidimensional
# shape: a tuple of integers
# i: integer used as an integer index
# idx: generic index (Python type)
# index: generic index (ndindex type)
# s: slice (Python type)
# S: Slice (ndindex type)
# size: integer passed to arange



from numpy import prod, arange, array, int8, intp, empty

from hypothesis import given, example
from hypothesis.strategies import one_of, integers

from pytest import raises

from .helpers import integer_arrays, short_shapes, check_same, assert_equal

from ..integer import Integer
from ..integerarray import IntegerArray

def test_integerarray_constructor():
    raises(ValueError, lambda: IntegerArray([0], shape=(1,)))
    raises(ValueError, lambda: IntegerArray([], shape=(1,)))
    raises(TypeError, lambda: IntegerArray([False]))
    raises(TypeError, lambda: IntegerArray(array(0.0)))
    raises(TypeError, lambda: IntegerArray((1,)))
    idx = IntegerArray(array([0], dtype=int8))
    assert_equal(idx.array, array([0], dtype=intp))

    idx = IntegerArray([], shape=(0, 1))
    assert_equal(idx.array, empty((0, 1), dtype=intp))

    # Make sure the underlying array is immutable
    idx = IntegerArray([1, 2])
    with raises(ValueError):
        idx.array[0] = 0
    assert_equal(idx.array, array([1, 2], dtype=intp))

    # Make sure the underlying array is copied
    a = array([1, 2])
    idx = IntegerArray(a)
    a[0] = 0
    assert idx == IntegerArray([1, 2])

@given(integer_arrays, short_shapes)
def test_integerarray_hypothesis(idx, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, idx)

@given(integer_arrays, one_of(short_shapes, integers(0, 10)))
def test_integerarray_reduce_no_shape_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = IntegerArray(idx)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.reduce().raw])

@example(array([2, 0]), (1, 0))
@example(array(0), 1)
@example(array([], dtype=intp), 0)
@given(integer_arrays, one_of(short_shapes, integers(0, 10)))
def test_integerarray_reduce_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = IntegerArray(idx)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.reduce(shape).raw])

    try:
        reduced = index.reduce(shape)
    except IndexError:
        pass
    else:
        if isinstance(reduced, Integer):
            assert reduced.raw >= 0
        else:
            assert isinstance(reduced, IntegerArray)
            assert (reduced.raw >= 0).all()

@example([], (1,))
@example([0], (1, 0))
@given(integer_arrays, one_of(short_shapes, integers(0, 10)))
def test_integerarray_isempty_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = IntegerArray(idx)


    def raw_func(a, idx):
        return a[idx].size == 0

    def ndindex_func(a, index):
        return index.isempty(), index.isempty(shape)

    def assert_equal(raw_empty, ndindex_empty):
        isempty, isempty_shape = ndindex_empty

        # If isempty is True then a[t] should be empty
        if isempty:
            assert raw_empty, (index, shape)
        # We cannot test the converse with hypothesis. isempty may be False
        # but a[idx] could still be empty for this specific a (e.g., if a is
        # already itself empty).

        # If isempty is true with no shape it should be true for a specific
        # shape. The converse is not true because the indexed array could be
        # empty.
        if isempty:
            assert isempty_shape, (index, shape)

        # isempty() should always give the correct result for a specific
        # array after reduction
        assert isempty_shape == raw_empty, (index, shape)

    check_same(a, idx, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



from .ndindex import NDIndex, asshape
from .tuple import Tuple

class ellipsis(NDIndex):
    """
    Represents an ellipsis index, i.e., `...` (or `Ellipsis`).

    Ellipsis indices by themselves return the full array. Inside of a tuple
    index, an ellipsis skips 0 or more axes of the array so that everything
    after the ellipsis indexes the last axes of the array. A tuple index can
    have at most one ellipsis.

    For example `a[(0, ..., -2)]` would index the first element on the first
    axis, the second-to-last element in the last axis, and include all the
    axes in between.

    >>> from numpy import arange
    >>> a = arange(2*3*4).reshape((2, 3, 4))
    >>> a
    array([[[ 0,  1,  2,  3],
            [ 4,  5,  6,  7],
            [ 8,  9, 10, 11]],
           [[12, 13, 14, 15],
            [16, 17, 18, 19],
            [20, 21, 22, 23]]])
    >>> a[0, ..., -2]
    array([ 2,  6, 10])

    An ellipsis can go at the beginning of end of a tuple index, and is
    allowed to match 0 axes.

    .. note::

       Unlike the standard Python `Ellipsis`, `ellipsis` is the type, not the
       object (the name is lowercase to avoid conflicting with the built-in).
       Use `ellipsis()` or `ndindex(...)` to create the object. In most
       ndindex contexts, `...` can be used instead of `ellipsis()`, for
       instance, when creating a `Tuple` object. Also unlike `Ellipsis`,
       `ellipsis()` is not singletonized, so you should not use `is` to
       compare it. See the document on :any:`type-confusion` for more details.

    """
    def _typecheck(self):
        return ()

    def reduce(self, shape=None):
        """
        Reduce an ellipsis index

        Since an ellipsis by itself always returns the full array unchanged,
        `ellipsis().reduce()` returns `Tuple()` as a canonical form (the index
        `()` also always returns an array unchanged).

        >>> from ndindex import ellipsis
        >>> ellipsis().reduce()
        Tuple()

        See Also
        ========

        .NDIndex.reduce
        .Tuple.reduce
        .Slice.reduce
        .Newaxis.reduce
        .Integer.reduce
        .IntegerArray.reduce
        .BooleanArray.reduce

        """
        if shape is not None:
            shape = asshape(shape)
        return Tuple()

    @property
    def raw(self):
        return ...

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        shape = asshape(shape)

        return shape

    def as_subindex(self, index):
        return Tuple().as_subindex(index)

    def isempty(self, shape=None):
        return Tuple().isempty(shape=shape)

    def __eq__(self, other):
        return other is ... or isinstance(other, ellipsis)

    def __hash__(self):
        return super().__hash__()



from numpy import intp, ndarray

from .array import ArrayIndex
from .ndindex import asshape
from .subindex_helpers import subindex_slice

class IntegerArray(ArrayIndex):
    """
    Represents an integer array index.

    If `idx` is an n-dimensional integer array with shape `s = (s1, ..., sn)`
    and `a` is any array, `a[idx]` replaces the first dimension of `a` with
    dimensions of size `s1, ..., sn`, where each entry is indexed according to
    the entry in `idx` as an integer index.

    Integer arrays can also appear as part of tuple indices. In that case,
    they replace the axis being indexed. If more than one integer array
    appears inside of a tuple index, they are broadcast together and iterated
    as one. Furthermore, if an integer array appears in a tuple index, all
    integer indices in the tuple are treated as scalar integer arrays and are
    also broadcast. In general, an :any:`Integer` index semantically behaves
    the same as a scalar (`shape=()`) `IntegerArray`.

    A list of integers may also be used in place of an integer array. Note
    that NumPy treats a direct list of integers as a tuple index, but this
    behavior is deprecated and will be replaced with integer array indexing in
    the future. ndindex always treats lists as arrays.

    >>> from ndindex import IntegerArray
    >>> import numpy as np
    >>> idx = IntegerArray([[0, 1], [1, 2]])
    >>> a = np.arange(10)
    >>> a[idx.raw]
    array([[0, 1],
           [1, 2]])

    .. note::

       `IntegerArray` does *not* represent an array, but rather an *array
       index*. It does not have most methods that `numpy.ndarray` has, and
       should not be used in array contexts. See the document on
       :any:`type-confusion` for more details.

    """
    dtype = intp
    """
    The dtype of `IntegerArray` is `np.intp`, which is typically either
    `np.int32` or `np.int64` depending on the platform.
    """

    def reduce(self, shape=None, axis=0):
        """
        Reduce an `IntegerArray` index on an array of shape `shape`.

        The result will either be `IndexError` if the index is invalid for the
        given shape, an `IntegerArray` index where the values are all
        nonnegative, or, if `self` is a scalar array index (`self.shape ==
        ()`), an `Integer` whose value is nonnegative.

        >>> from ndindex import IntegerArray
        >>> idx = IntegerArray([-5, 2])
        >>> idx.reduce((3,))
        Traceback (most recent call last):
        ...
        IndexError: index -5 is out of bounds for axis 0 with size 3
        >>> idx.reduce((9,))
        IntegerArray([4, 2])

        See Also
        ========

        .NDIndex.reduce
        .Tuple.reduce
        .Slice.reduce
        .ellipsis.reduce
        .Newaxis.reduce
        .Integer.reduce
        .BooleanArray.reduce

        """
        from .integer import Integer

        if self.shape == ():
            return Integer(self.array).reduce(shape, axis=axis)

        if shape is None:
            return self

        shape = asshape(shape, axis=axis)

        size = shape[axis]
        new_array = self.array.copy()
        out_of_bounds = (new_array >= size) | ((-size > new_array) & (new_array < 0))
        if out_of_bounds.any():
            raise IndexError(f"index {new_array[out_of_bounds].flat[0]} is out of bounds for axis {axis} with size {size}")

        new_array[new_array < 0] += size
        return IntegerArray(new_array)

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        shape = asshape(shape)

        # reduce will raise IndexError if it should be raised
        self.reduce(shape)
        return self.shape + shape[1:]

    def isempty(self, shape=None):
        if shape is not None:
            return 0 in self.newshape(shape)

        return 0 in self.shape

    def as_subindex(self, index):
        from .ndindex import ndindex
        from .slice import Slice
        from .tuple import Tuple

        index = ndindex(index)

        if isinstance(index, Tuple):
            return Tuple(self).as_subindex(index)

        if isinstance(index, Slice):
            index = index.reduce()
            if (self.array < 0).any():
                raise NotImplementedError("IntegerArray.as_subindex() is only implemented for arrays with all nonnegative entries. Try calling reduce() with a shape first.")
            if index.step < 0:
                raise NotImplementedError("IntegerArray.as_subindex(Slice) is only implemented for slices with positive steps")

            # After reducing, start is not None when step > 0
            if index.stop is None or index.start < 0 or index.stop < 0:
                raise NotImplementedError("IntegerArray.as_subindex(Slice) is only implemented for slices with nonnegative start and stop. Try calling reduce() with a shape first.")

            # Equivalent to

            # res = []
            # for i in self.array.flat:
            #     s = Slice(i, i+1).as_subindex(index)
            #     if s == Slice(0, 0, 1):
            #         continue
            #     res.append(s.start)
            # return IntegerArray(res)

            # See also Integer.as_subindex().

            s = self.array
            start, stop, step = subindex_slice(s, s+1, 1,
                                               index.start, index.stop, index.step)
            if (stop <= 0).all():
                raise ValueError("Indices do not intersect")

            if start.shape == ():
                if start < stop:
                    return IntegerArray(start)
                else:
                    raise ValueError("Indices do not intersect")

            mask = start < stop
            if not mask.all():
                start = start[start < stop]

            if 0 in start.shape:
                raise ValueError("Indices do not intersect")
            return IntegerArray(start)

        raise NotImplementedError("IntegerArray.as_subindex is only implemented for slices")

    def __eq__(self, other):
        if isinstance(other, IntegerArray):
            b = other.array
        elif isinstance(other, ndarray):
            b = other
        elif isinstance(other, list):
            try:
                b = IntegerArray(other)
            except TypeError:
                return False
        else:
            return False
        a = self.array
        return a.shape == b.shape and (a == b).all()

    def __hash__(self):
        return super().__hash__()



from .ndindex import NDIndex, asshape, operator_index
from .subindex_helpers import subindex_slice

class default:
    """
    A default keyword argument value.

    Used as the default value for keyword arguments where `None` is also a
    meaningful value but not the default.

    """
    pass

class Slice(NDIndex):
    """
    Represents a slice on an axis of an nd-array.

    `Slice(x)` with one argument is equivalent to `Slice(None, x)`. `Slice(x,
    y)` with two arguments is equivalent to `Slice(x, y, None)`.

    `start` and `stop` can be any integer, or `None`. `step` can be any
    nonzero integer or `None`.

    `Slice(a, b)` is the same as the syntax `a:b` in an index and `Slice(a, b,
    c)` is the same as `a:b:c`. An argument being `None` is equivalent to the
    syntax where the item is omitted, for example, `Slice(None, None, k)` is
    the same as the syntax `::k`.

    `Slice.args` always has three arguments, and does not make any distinction
    between, for instance, `Slice(x, y)` and `Slice(x, y, None)`. This is
    because Python itself does not make the distinction between x:y and x:y:
    syntactically.

    See :ref:`slices-docs` for a description of the semantic meaning of slices
    on arrays.

    Slice has attributes `start`, `stop`, and `step` to access the
    corresponding attributes.

    >>> from ndindex import Slice
    >>> s = Slice(10)
    >>> s
    Slice(None, 10, None)
    >>> print(s.start)
    None
    >>> s.args
    (None, 10, None)
    >>> s.raw
    slice(None, 10, None)

    """
    def _typecheck(self, start, stop=default, step=None):
        if isinstance(start, Slice):
            return start.args
        if isinstance(start, slice):
            start, stop, step = start.start, start.stop, start.step

        if stop is default:
            start, stop = None, start

        if step == 0:
            raise ValueError("slice step cannot be zero")

        if start is not None:
            start = operator_index(start)
        if stop is not None:
            stop = operator_index(stop)
        if step is not None:
            step = operator_index(step)

        args = (start, stop, step)

        return args

    def __hash__(self):
        # We can't use the default hash(self.raw) because slices are not
        # hashable
        return hash(self.args)

    @property
    def raw(self):
        return slice(*self.args)

    @property
    def start(self):
        """
        The start value of the slice.

        Note that this may be an integer or `None`.
        """
        return self.args[0]

    @property
    def stop(self):
        """
        The stop of the slice.

        Note that this may be an integer or `None`.
        """
        return self.args[1]

    @property
    def step(self):
        """
        The step of the slice.

        Note that this may be a nonzero integer or `None`.
        """
        return self.args[2]

    def __len__(self):
        """
        `len()` gives the maximum size of an axis sliced with `self`.

        An actual array may produce a smaller size if it is smaller than the
        bounds of the slice. For instance, `[0, 1, 2][2:4]` only has 1 element
        but the maximum length of the slice `2:4` is 2.

        >>> from ndindex import Slice
        >>> [0, 1, 2][2:4]
        [2]
        >>> len(Slice(2, 4))
        2
        >>> [0, 1, 2, 3][2:4]
        [2, 3]

        If there is no such maximum, it raises `ValueError`.

        >>> # From the second element to the end, which could have any size
        >>> len(Slice(1, None))
        Traceback (most recent call last):
        ...
        ValueError: Cannot determine max length of slice

        The :meth:`Slice.reduce` method with a `shape` argument returns a
        `Slice` that always has a correct `len` which doesn't raise
        `ValueError`.

        >>> Slice(2, 4).reduce(3)
        Slice(2, 3, 1)
        >>> len(_)
        1

        Be aware that `len(Slice)` only gives the size of the axis being
        sliced. It does not say anything about the total shape of the array.
        In particular, the array may be empty after slicing if one of its
        dimensions is 0, but the other dimensions may be nonzero. To check if
        an array will empty after indexing, use :meth:`isempty`.

        See Also
        ========
        isempty

        """
        start, stop, step = self.reduce().args
        error = ValueError("Cannot determine max length of slice")
        # We reuse the logic in range.__len__. However, it is only correct if
        # start and stop are nonnegative.
        if step > 0:
            # start cannot be None
            if stop is None:
                if start >= 0:
                    # a[n:]. Extends to the end of the array.
                    raise error
                else:
                    # a[-n:]. From n from the end to the end. Same as
                    # range(-n, 0).
                    stop = 0
            elif start < 0 and stop >= 0:
                # a[-n:m] indexes from nth element from the end to the
                # m-1th element from the beginning.
                start, stop = 0, min(-start, stop)
            elif start >=0 and stop < 0:
                # a[n:-m]. The max length depends on the size of the array.
                raise error
        else:
            if stop is None:
                if start >= 0:
                    # a[n::-1] (start != None by above). Same as range(n, -1, -1)
                    stop = -1
                else:
                    # a[-n::-1]. From n from the end to the beginning of the
                    # array backwards. The max length depends on the size of
                    # the array.
                    raise error
            elif start < 0 and stop >= 0:
                # a[-n:m:-1]. The max length depends on the size of the array
                raise error
            elif start >=0 and stop < 0:
                # a[n:-m:-1] indexes from the nth element backwards to the mth
                # element from the end.
                start, stop = 0, min(start+1, -stop - 1)
                step = -step

        return len(range(start, stop, step))

    def reduce(self, shape=None, axis=0):
        """
        `Slice.reduce` returns a slice that is canonicalized for an array of the
        given shape, or for any shape if `shape` is `None` (the default).

        `Slice.reduce` is a perfect canonicalization, meaning that two slices
        are equal---for all array shapes if `shape=None` or for arrays of shape
        `shape` otherwise---if and only if they `reduce` to the same Slice
        object. Note that ndindex objects do not simplify automatically, and
        `==` only does exact equality comparison, so to test that two slices
        are equal, use `slice1.reduce(shape) == slice2.reduce(shape)`.

        - If `shape` is `None`, the following properties hold after calling
          `reduce()`:

          - `start` is not `None`.

          - `stop` is not `None`, when possible. The reduced `stop` can only
            be `None` if the original `stop` is.

          - `step` is not `None`.

          - `step` is as close to 0 as possible.

          - If the slice is always empty, the resulting slice will be
            `Slice(0, 0, 1)`. However, one should prefer the :any:`isempty`
            method to test if a slice is always empty.

          In particular, `stop` may be `None`, even after canonicalization
          with `reduce()` with no `shape`. This is because some slices are
          impossible to represent without `None` without making assumptions
          about the array shape. For example, `Slice(0, None)` cannot be
          equivalent to a slice with `stop != None` for all array shapes. To
          get a slice where the `start`, `stop`, and `step` are always
          integers, use `reduce(shape)` with an explicit array shape.

          Note that `Slice` objects that index a single element are not
          canonicalized to `Integer`, because integer indices always remove an
          axis whereas slices keep the axis. Furthermore, slices cannot raise
          `IndexError` except on arrays with shape equal to `()`.

          >>> from ndindex import Slice
          >>> Slice(10).reduce()
          Slice(0, 10, 1)
          >>> Slice(1, 3, 3).reduce()
          Slice(1, 2, 1)

        - If an explicit shape is given, the following properties are true
          after calling `Slice.reduce(shape)`:

          - `start`, `stop`, and `step` are not `None`,

          - `start` is nonnegative.

          - `stop` is nonnegative whenever possible. In particular, `stop` is
            only negative when it has to be to represent the given slice,
            i.e., a slice with negative `step` that indexes more than 1
            element and indexes the first (index `0`) element (in this case,
            it will be `-n - 1` where `n` is the size of the axis being
            sliced).

          - `stop` is as small as possible for positive `step` or large as
            possible for negative `step`.

          - `step` is as close to 0 as possible.

          - If the slice is empty for the given shape, the resulting slice
            will be `Slice(0, 0, 1)`. However, one should prefer the
            :any:`isempty` method to test if a slice is always empty.

          - If the slice indexes a single element, the resulting slice will be
            of the form `Slice(i, i+1, 1)`. However, one should prefer using
            `len(s.reduce(shape)) == 1` to test if a slice indexes exactly 1
            element.

          - :any:`len() <Slice.__len__>` gives the true size of the axis for a
            sliced array of the given shape, and never raises `ValueError`.

          The `axis` argument can be used to specify an axis of the shape (by
          default, `axis=0`). For convenience, `shape` can be passed as an integer
          for a single dimension.


          >>> from ndindex import Slice
          >>> Slice(1, 10).reduce(3)
          Slice(1, 3, 1)
          >>> Slice(-1, 1, -2).reduce(4)
          Slice(3, 4, 1)
          >>> Slice(1, 10, 3).reduce((4, 5), axis=0)
          Slice(1, 2, 1)
          >>> Slice(1, 10, 3).reduce((4, 5), axis=1)
          Slice(1, 5, 3)

          >>> s = Slice(2, None)
          >>> len(s)
          Traceback (most recent call last):
          ...
          ValueError: Cannot determine max length of slice
          >>> s.reduce((5,))
          Slice(2, 5, 1)
          >>> len(_)
          3

        See Also
        ========

        .NDIndex.reduce
        .Tuple.reduce
        .Integer.reduce
        .ellipsis.reduce
        .Newaxis.reduce
        .IntegerArray.reduce
        .BooleanArray.reduce

        """
        start, stop, step = self.args

        # Canonicalize with no shape

        if step is None:
            step = 1
        if start is None:
            if step > 0:
                start = 0
            else: # step < 0
                start = -1

        if start is not None and stop is not None:
            if start >= 0 and stop >= 0 or start < 0 and stop < 0:
                if step > 0:
                    if stop <= start:
                        start, stop, step = 0, 0, 1
                    elif start >= 0 and start + step >= stop:
                        # Indexes 1 element. Start has to be >= 0 because a
                        # negative start could be less than the size of the
                        # axis, in which case it will clip and the single
                        # element will be element 0. We can only do that
                        # reduction if we know the shape.

                        # Note that returning Integer here is wrong, because
                        # slices keep the axis and integers remove it.
                        stop, step = start + 1, 1
                    elif start < 0 and start + step > stop:
                        # The exception is this case where stop is already
                        # start + 1.
                        step = stop - start
                    if start >= 0:
                        stop -= (stop - start - 1) % step
                else: # step < 0
                    if stop >= start:
                        start, stop, step = 0, 0, 1
                    elif start < 0 and start + step <= stop:
                        if start < -1:
                            stop, step = start + 1, 1
                        else: # start == -1
                            stop, step = start - 1, -1
                    elif stop == start - 1:
                        stop, step = start + 1, 1
                    elif start >= 0 and start + step <= stop:
                        # Indexes 0 or 1 elements. We can't change stop
                        # because start might clip to a smaller true start if
                        # the axis is smaller than it, and increasing stop
                        # would prevent it from indexing an element in that
                        # case. The exception is the case right before this
                        # one (stop == start - 1). In that case start cannot
                        # clip past the stop (it always indexes the same one
                        # element in the cases where it indexes anything at
                        # all).
                        step = stop - start
                    if start < 0:
                        stop -= (stop - start + 1) % step
            elif start >= 0 and stop < 0 and step < 0 and (start < -step or
                                                           -stop - 1 < -step):
                if stop == -1:
                    start, stop, step = 0, 0, 1
                else:
                    step = max(-start - 1, stop + 1)
            elif start < 0 and stop == 0 and step > 0:
                start, stop, step = 0, 0, 1
            elif start < 0 and stop >= 0 and step >= min(-start, stop):
                step = min(-start, stop)
                if start == -1 or stop == 1:
                    # Can only index 0 or 1 elements. We can either pick a
                    # version with positive start and negative step, or
                    # negative start and positive step. We prefer the former
                    # as it matches what is done for reduce() with a shape
                    # (start is always nonnegative).
                    assert step == 1
                    start, stop, step = stop - 1, start - 1, -1
        elif start is not None and stop is None:
            if start == -1 and step > 0:
                start, stop, step = (-1, -2, -1)
            elif start < 0 and step >= -start:
                step = -start
            elif step < 0:
                if start == 0:
                    start, stop, step = 0, 1, 1
                elif 0 <= start < -step:
                    step = -start - 1
        if shape is None:
            return type(self)(start, stop, step)

        # Further canonicalize with an explicit array shape

        shape = asshape(shape, axis=axis)
        size = shape[axis]

        if stop is None:
            if step > 0:
                stop = size
            else:
                stop = -size - 1

        if stop < -size:
            stop = -size - 1

        if size == 0:
            start, stop, step = 0, 0, 1
        elif step > 0:
            # start cannot be None
            if start < 0:
                start = size + start
            if start < 0:
                start = 0
            if start >= size:
                start, stop, step = 0, 0, 1

            if stop < 0:
                stop = size + stop
                if stop < 0:
                    stop = 0
            else:
                stop = min(stop, size)
            stop -= (stop - start - 1) % step

            if stop - start == 1:
                # Indexes 1 element.
                step = 1
            elif stop - start <= 0:
                start, stop, step = 0, 0, 1
        else:
            if start < 0:
                if start >= -size:
                    start = size + start
                else:
                    start, stop = 0, 0
            if start >= 0:
                start = min(size - 1, start)

            if -size <= stop < 0:
                stop += size

            if stop >= 0:
                if start - stop == 1:
                    stop, step = start + 1, 1
                elif start - stop <= 0:
                    start, stop, step = 0, 0, 1
                else:
                    stop += (start - stop - 1) % -step

            # start >= 0
            if (stop < 0 and start - size - stop <= -step
                or stop >= 0 and start - stop <= -step):
                stop, step = start + 1, 1
            if stop < 0 and start % step != 0:
                # At this point, negative stop is only necessary to index the
                # first element. If that element isn't actually indexed, we
                # prefer a nonnegative stop. Otherwise, stop will be -size - 1.
                stop = start % -step - 1
        return self.__class__(start, stop, step)

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        shape = asshape(shape)

        idx = self.reduce(shape)

        # len() won't raise an error after reducing with a shape
        return (len(idx),) + shape[1:]

    # TODO: Better name?
    def as_subindex(self, index):
        # The docstring of this method is currently on NDindex.as_subindex, as
        # this is the only method that is actually implemented so far.

        from .ndindex import ndindex
        from .tuple import Tuple
        from .integer import Integer
        from .integerarray import IntegerArray
        from .booleanarray import BooleanArray

        index = ndindex(index)
        index_orig = index

        s = self.reduce()
        index = index.reduce()

        if isinstance(index, Tuple):
            return Tuple(self).as_subindex(index)

        if isinstance(index, Integer):
            if index == -1:
                s = self.as_subindex(Slice(index.args[0], None))
            else:
                s = self.as_subindex(Slice(index.args[0], index.args[0] + 1))
            if s == Slice(0, 0, 1):
                # There is no index that we can return here. The intersection
                # of `self` and `index` is empty. Ideally we want to give an
                # index that gives an empty array, but we cannot make the
                # shape match. If a is dimension 1, then a[index] is dimension
                # 0, so a[index][slice(0, 0)] will not work. A possibility
                # would be to return False, which would add a length-0
                # dimension to the array. But
                #
                # 1. this isn't implemented yet, and
                # 2. a False can only add a length-0 dimension once, so it
                #    still wouldn't work in every case. For example,
                #    Tuple(slice(0), slice(0)).as_subindex((0, 0)) would need
                #    to return an index that replaces the first two
                #    dimensions with length-0 dimensions.
                raise ValueError(f"{self} and {index_orig} do not intersect")
            assert len(s) == 1
            return Tuple()

        if s.step < 0:
            raise NotImplementedError("Slice.as_subindex() is only implemented for slices with positive steps")

        # After reducing, start is not None when step > 0
        if s.stop is None or s.start < 0 or s.stop < 0:
            raise NotImplementedError("Slice.as_subindex() is only implemented for slices with nonnegative start and stop. Try calling reduce() with a shape first.")

        if isinstance(index, IntegerArray):
            idx = index.array
            if (idx < 0).any():
                raise NotImplementedError("Slice.as_subindex(IntegerArray) is not yet implemented for arrays with negative values. Try calling reduce with a shape first.")
            start, stop, step = subindex_slice(s.start, s.stop, s.step,
                                               idx, idx+1, 1)
            res = BooleanArray(start < stop)

            if not res.count_nonzero:
                raise ValueError("Indices do not intersect")

            return res

        if not isinstance(index, Slice):
            raise NotImplementedError("Slice.as_subindex() is only implemented for tuples, integers, arrays and slices")

        if index.step < 0:
            raise NotImplementedError("Slice.as_subindex() is only implemented for slices with positive steps")

        # After reducing, start is not None when step > 0
        if index.stop is None or index.start < 0 or index.stop < 0:
            raise NotImplementedError("Slice.as_subindex() is only implemented for slices with nonnegative start and stop. Try calling reduce() with a shape first.")

        return Slice(*subindex_slice(s.start, s.stop, s.step, index.start,
                                     index.stop, index.step)).reduce()

    def isempty(self, shape=None):
        if shape is not None:
            return 0 in self.newshape(shape)

        try:
            l = len(self)
        except (TypeError, ValueError):
            return False
        return l == 0

    def __eq__(self, other):
        if isinstance(other, slice):
            return self.args == (other.start, other.stop, other.step)
        elif isinstance(other, Slice):
            return self.args == other.args
        return False



from numpy import arange, newaxis

from hypothesis import given
from hypothesis.strategies import one_of, integers

from ..ndindex import ndindex
from .helpers import check_same, prod, shapes, newaxes

def test_newaxis_exhaustive():
    for n in range(10):
        a = arange(n)
    check_same(a, newaxis)


@given(newaxes(), shapes)
def test_newaxis_hypothesis(idx, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, idx)


def test_newaxis_reduce_exhaustive():
    for n in range(10):
        a = arange(n)
        check_same(a, newaxis, ndindex_func=lambda a, x: a[x.reduce((n,)).raw])


@given(newaxes(), shapes)
def test_newaxis_reduce_hypothesis(idx, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, idx, ndindex_func=lambda a, x: a[x.reduce(shape).raw])


def test_newaxis_reduce_no_shape_exhaustive():
    for n in range(10):
        a = arange(n)
        check_same(a, newaxis, ndindex_func=lambda a, x: a[x.reduce().raw])

@given(newaxes(), shapes)
def test_newaxis_reduce_no_shape_hypothesis(idx, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, idx, ndindex_func=lambda a, x: a[x.reduce().raw])

@given(newaxes(), one_of(shapes, integers(0, 10)))
def test_newaxis_isempty_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = ndindex(idx)


    def raw_func(a, idx):
        return a[idx].size == 0

    def ndindex_func(a, index):
        return index.isempty(), index.isempty(shape)

    def assert_equal(raw_empty, ndindex_empty):
        isempty, isempty_shape = ndindex_empty

        # Since idx is a newaxis, it should never be unconditionally empty
        assert not isempty
        # We cannot test the converse with hypothesis. isempty may be False
        # but a[idx] could still be empty for this specific a (e.g., if a is
        # already itself empty).

        # isempty() should always give the correct result for a specific
        # array after reduction
        assert isempty_shape == raw_empty, (index, shape)

    check_same(a, idx, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



import inspect
import numbers

from numpy import ndarray, bool_, newaxis

def ndindex(obj):
    """
    Convert an object into an ndindex type

    Invalid indices will raise `IndexError`.

    >>> from ndindex import ndindex
    >>> ndindex(1)
    Integer(1)
    >>> ndindex(slice(0, 10))
    Slice(0, 10, None)

    """
    if isinstance(obj, NDIndex):
        return obj

    if isinstance(obj, (bool, bool_)):
        from . import BooleanArray
        return BooleanArray(obj)

    if isinstance(obj, (list, ndarray)):
        from . import IntegerArray, BooleanArray

        try:
            return IntegerArray(obj)
        except TypeError:
            pass
        try:
            return BooleanArray(obj)
        except TypeError:
            pass

        # Match the NumPy exceptions
        if isinstance(obj, ndarray):
            raise IndexError("arrays used as indices must be of integer (or boolean) type")
        else:
            raise IndexError("only integers, slices (`:`), ellipsis (`...`), numpy.newaxis (`None`) and integer or boolean arrays are valid indices")

    try:
        from . import Integer
        # If operator.index() works, use that
        return Integer(obj)
    except TypeError:
        pass

    if isinstance(obj, slice):
        from . import Slice
        return Slice(obj)

    if isinstance(obj, tuple):
        from . import Tuple
        return Tuple(*obj)

    from . import ellipsis

    if obj == ellipsis:
        raise IndexError("Got ellipsis class. Did you mean to use the instance, ellipsis()?")
    if obj is Ellipsis:
        return ellipsis()

    if obj == newaxis:
        from . import Newaxis
        return Newaxis()

    raise IndexError("only integers, slices (`:`), ellipsis (`...`), numpy.newaxis (`None`) and integer or boolean arrays are valid indices")

class classproperty(object):
    def __init__(self, f):
        self.f = f
    def __get__(self, obj, owner):
        return self.f(owner)

class ImmutableObject:
    """
    Base class for immutable objects.

    Subclasses of this class are immutable objects. They all have the `.args`
    attribute, which gives the full necessary data to recreate the class, via,

    .. code:: python

       type(obj)(*obj.args) == obj

    Note: subclasses that specifically represent indices should subclass
    :class:`NDIndex` instead.

    All classes that subclass `ImmutableObject` should define the `_typecheck`
    method. `_typecheck(self, *args)` should do type checking and basic type
    canonicalization, and either return a tuple of the new arguments for the
    class or raise an exception. Type checking means it should raise
    exceptions for input types that are never semantically meaningful for
    numpy arrays, for example, floating point indices, using the same
    exceptions as numpy where possible. Basic type canonicalization means, for
    instance, converting integers into `int` using `operator.index()`. All
    other canonicalization should be done in the `reduce()` method. The
    `ImmutableObject` base constructor will automatically set `.args` to the
    arguments returned by this method. Classes should always be able to
    recreate themselves with `.args`, i.e., `type(obj)(*obj.args) == obj`
    should always hold.

    See Also
    ========

    NDIndex

    """
    def __init__(self, *args, **kwargs):
        """
        This method should be called by subclasses (via super()) after type-checking
        """
        args = self._typecheck(*args, **kwargs)
        self.args = args
        """
        `idx.args` contains the arguments needed to create `idx`.

        For an ndindex object `idx`, `idx.args` is always a tuple such that

        .. code:: python

           type(idx)(*idx.args) == idx

        For :any:`Tuple` indices, the elements
        of `.args` are themselves ndindex types. For other types, `.args`
        contains raw Python types. Note that `.args` contains NumPy arrays for
        :any:`IntegerArray` and :any:`BooleanArray` types, so one should
        always do equality testing or hashing on the ndindex type itself, not
        its `.args`.
        """
    @classproperty
    def __signature__(self):
        """
        Allow Python 3's inspect.signature to give a useful signature for
        NDIndex subclasses.
        """
        sig = inspect.signature(self._typecheck)
        d = dict(sig.parameters)
        d.pop('self')
        return inspect.Signature(d.values())

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(map(repr, self.args))})"

    def __str__(self):
        return f"{self.__class__.__name__}({', '.join(map(str, self.args))})"

    def __eq__(self, other):
        if not isinstance(other, ImmutableObject):
            try:
                other = self.__class__(other)
            except TypeError:
                return False

        return self.args == other.args

    def __hash__(self): # pragma: no cover
        # Note: subclasses where .args is not hashable should redefine
        # __hash__
        return hash(self.args)

class NDIndex(ImmutableObject):
    """
    Represents an index into an nd-array (i.e., a numpy array).

    This is a base class for all ndindex types. All types that subclass this
    class should redefine the following methods

    - `_typecheck(self, *args)`. See the docstring of
      :class:`ImmutableObject`.

    - `raw` (a **@property** method) should return the raw index that can be
      passed as an index to a numpy array.

    In addition other methods should be defined as necessary.

    - `__len__` should return the largest possible shape of an axis sliced by
      the index (for single-axis indices), or raise ValueError if no such
      maximum exists.

    - `reduce(shape=None)` should reduce an index to an equivalent form for
      arrays of shape `shape`, or raise an `IndexError`. The error messages
      should match numpy as much as possible. The class of the equivalent
      index may be different. If `shape` is `None`, it should return a
      canonical form that is equivalent for all array shapes (assuming no
      IndexErrors).

    The methods `__init__` and `__eq__` should *not* be overridden. Equality
    (and hashability) on `NDIndex` subclasses is determined by equality of
    types and `.args`. Equivalent indices should not attempt to redefine
    equality. Rather they should define canonicalization via `reduce()`.
    `__hash__` is defined so that the hash matches the hash of `.raw`. If
    `.raw` is unhashable, `__hash__` should be overridden to use
    `hash(self.args)`.

    See Also
    ========

    ImmutableObject

    """
    # TODO: Make NDIndex and ImmutableObject abstract base classes
    @property
    def raw(self):
        """
        Return the equivalent of `self` that can be used as an index

        NumPy does not allow custom objects to be used as indices, with the
        exception of integer indices, so to use an ndindex object as an
        index, it is necessary to use `raw`.

        >>> from ndindex import Slice
        >>> import numpy as np
        >>> a = np.arange(5)
        >>> s = Slice(2, 4)
        >>> a[s]
        Traceback (most recent call last):
        ...
        IndexError: only integers, slices (`:`), ellipsis (`...`), numpy.newaxis (`None`) and integer or boolean arrays are valid indices
        >>> a[s.raw]
        array([2, 3])

        """
        raise NotImplementedError

    # This is still here as a fallback implementation, but it isn't actually
    # used by any present subclasses, because it is faster to implement __eq__
    # on each class specifically.
    def __eq__(self, other): # pragma: no cover
        if not isinstance(other, NDIndex):
            try:
                other = ndindex(other)
                return self == other
            except IndexError:
                return False
        return super().__eq__(other)

    def __hash__(self):
        # Make the hash match the raw hash when the raw type is hashable.
        # Note: subclasses where .raw is not hashable should define __hash__
        # as hash(self.args)
        return hash(self.raw)

    def reduce(self, shape=None):
        """
        Simplify an index given that it will be applied to an array of a given shape.

        If `shape` is None (the default), the index will be canonicalized as
        much as possible while still staying equivalent for all array shapes
        that it does not raise IndexError for.

        Either returns a new index type, which is equivalent on arrays of
        shape `shape`, or raises IndexError if the index would give an index
        error (for instance, out of bounds integer index or too many indices
        for array).

        >>> from ndindex import Slice, Integer
        >>> Slice(0, 10).reduce((5,))
        Slice(0, 5, 1)
        >>> Integer(10).reduce((5,))
        Traceback (most recent call last):
        ...
        IndexError: index 10 is out of bounds for axis 0 with size 5

        For single axis indices such as Slice and Tuple, `reduce` takes an
        optional `axis` argument to specify the axis, defaulting to 0.

        See Also
        ========

        .Integer.reduce
        .Tuple.reduce
        .Slice.reduce
        .ellipsis.reduce
        .Newaxis.reduce
        .IntegerArray.reduce
        .BooleanArray.reduce

        """
        # XXX: Should the default be raise NotImplementedError or return self?
        raise NotImplementedError

    def expand(self, shape):
        r"""
        Expand a Tuple index on an array of shape `shape`

        An expanded index is as explicit as possible. Unlike :any:`reduce
        <NDIndex.reduce>`, which tries to simplify an index and remove
        redundancies, `expand()` typically makes an index larger.

        If `self` is invalid for the given shape, an `IndexError` is raised.
        Otherwise, the returned index satisfies the following:

        - It is always a :any:`Tuple`.

        - All the elements of the :any:`Tuple` are recursively :any:`reduced
          <NDIndex.reduce>`.

        - The length of the `.args` is equal to the length of the shape plus
          the number of :any:`Newaxis` indices in `self` plus 1 if there is a
          scalar :any:`BooleanArray` (`True` or `False`).

        - The resulting :any:`Tuple` has no :any:`ellipses <ellipsis>`. If
          there are axes that would be matched by an ellipsis or an implicit
          ellipsis at the end of the tuple, `Slice(0, n, 1)` indices are
          inserted, where `n` is the corresponding axis of the `shape`.

        - Any array indices in `self` are broadcast together. If `self`
          contains array indices (:any:`IntegerArray` or :any:`BooleanArray`),
          then any :any:`Integer` indices are converted into
          :any:`IntegerArray` indices of shape `()` and broadcast. Note that
          broadcasting is done in a memory efficient way so that even if the
          broadcasted shape is large it will not take up more memory than the
          original.

        - Scalar :any:`BooleanArray` arguments (`True` or `False`) are
          combined into a single term (the same as with :any:`Tuple.reduce`).

        - Non-scalar :any:`BooleanArray`\ s are all converted into equivalent
          :any:`IntegerArray`\ s via `nonzero()` and broadcast.

        >>> from ndindex import Tuple, Slice
        >>> Slice(None).expand((2, 3))
        Tuple(slice(0, 2, 1), slice(0, 3, 1))

        >>> idx = Tuple(slice(0, 10), ..., None, -3)
        >>> idx.expand((5, 3))
        Tuple(slice(0, 5, 1), None, 0)
        >>> idx.expand((1, 2, 3))
        Tuple(slice(0, 1, 1), slice(0, 2, 1), None, 0)
        >>> idx.expand((5,))
        Traceback (most recent call last):
        ...
        IndexError: too many indices for array: array is 1-dimensional, but 2 were indexed
        >>> idx.expand((5, 2))
        Traceback (most recent call last):
        ...
        IndexError: index -3 is out of bounds for axis 1 with size 2

        >>> idx = Tuple(..., [0, 1], -1)
        >>> idx.expand((1, 2, 3))
        Tuple(slice(0, 1, 1), [0, 1], [2, 2])

        See Also
        ========

        .Tuple.reduce
        broadcast_arrays

        """
        from .tuple import Tuple

        return Tuple(self).expand(shape)

    def newshape(self, shape):
        """
        Returns the shape of `a[idx.raw]`, assuming `a` has shape `shape`.

        `shape` should be a tuple of ints, or an int, which is equivalent to a
        1-D shape.

        Raises `IndexError` if `self` would be out of shape for an array of
        shape `shape`.

        >>> from ndindex import Slice, Integer, Tuple
        >>> shape = (6, 7, 8)
        >>> Integer(1).newshape(shape)
        (7, 8)
        >>> Integer(10).newshape(shape)
        Traceback (most recent call last):
        ...
        IndexError: index 10 is out of bounds for axis 0 with size 6
        >>> Slice(2, 5).newshape(shape)
        (3, 7, 8)
        >>> Tuple(0, ..., Slice(1, 3)).newshape(shape)
        (7, 2)

        """
        raise NotImplementedError

    def as_subindex(self, index):
        """
        `i.as_subindex(j)` produces an index `k` such that `a[j][k]` gives all of
        the elements of `a[j]` that are also in `a[i]`.

        If `a[j]` is a subset of `a[i]`, then `a[j][k] == a[i]`. Otherwise,
        `a[j][k] == a[i & j]`, where `i & j` is the intersection of `i` and
        `j`, that is, the elements of `a` that are indexed by both `i` and
        `j`.

        For example, in the below diagram, `i` and `j` index a subset of the
        array `a`. `k = i.as_subindex(j)` is an index on `a[j]` that gives the
        subset of `a[j]` also included in `a[i]`::

             +------------ self ------------+
             |                              |
         ------------------- a -----------------------
                |                                 |
                +------------- index -------------+
                |                           |
                +- self.as_subindex(index) -+

        `i.as_subindex(j)` is currently only implemented when `j` is a slices
        with positive steps and nonnegative start and stop, or a Tuple of the
        same. To use it with slices with negative start or stop, call
        :meth:`reduce` with a shape first.

        `as_subindex` can be seen as the left-inverse of composition, that is,
        if `a[i] = a[j][k]`, then `k = i.as_subindex(j)`, so that `k "="
        (j^-1)[i]` (this only works as a true inverse if `j` is a subset of
        `i`).

        Note that due to symmetry, `a[j][i.as_subindex(j)]` and
        `a[i][j.as_subindex(i)]` will give the same subarrays of `a`, which
        will be the array of elements indexed by both `a[i]` and `a[j]`.

        `i.as_subindex(j)` may raise `ValueError` in the case that the indices
        `i` and `j` do not intersect at all.

        Examples
        ========

        An example usage of `as_subindex` is to split an index up into
        subindices of chunks of an array. For example, say a 1-D array
        `a` is chunked up into chunks of size `N`, so that `a[0:N]`,
        `a[N:2*N]`, `[2*N:3*N]`, etc. are stored separately. Then an index
        `a[i]` can be reindexed onto the chunks via `i.as_subindex(Slice(0,
        N))`, `i.as_subindex(Slice(N, 2*N))`, etc.

        >>> from ndindex import Slice
        >>> i = Slice(5, 15)
        >>> j1 = Slice(0, 10)
        >>> j2 = Slice(10, 20)
        >>> a = list(range(20))
        >>> a[i.raw]
        [5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        >>> a[j1.raw]
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        >>> a[j2.raw]
        [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]

        >>> k1 = i.as_subindex(j1)
        >>> k1
        Slice(5, 10, 1)
        >>> k2 = i.as_subindex(j2)
        >>> k2
        Slice(0, 5, 1)
        >>> a[j1.raw][k1.raw]
        [5, 6, 7, 8, 9]
        >>> a[j2.raw][k2.raw]
        [10, 11, 12, 13, 14]

        See Also
        ========

        ndindex.ChunkSize.as_subchunks:
            a high-level iterator that efficiently gives only those chunks
            that intersect with a given index

        """
        index = ndindex(index) # pragma: no cover
        raise NotImplementedError(f"{type(self).__name__}.as_subindex({type(index).__name__}) isn't implemented yet")

    def isempty(self, shape=None):
        """
        Returns whether self always indexes an empty array

        An empty array is an array whose shape contains at least one 0. Note
        that scalars (arrays with shape `()`) are not considered empty.

        `shape` can be `None` (the default), or an array shape. If it is
        `None`, isempty() will return `True` when `self` is always empty for
        any array shape. However, if it gives `False`, it could still give an
        empty array for some array shapes, but not all. If you know the shape
        of the array that will be indexed, you can call `idx.isempty(shape)`
        first and the result will be correct for arrays of shape `shape`. If
        `shape` is given and `self` would raise an `IndexError` on an array of
        shape `shape`, `isempty()` also raises `IndexError`.

        >>> from ndindex import Tuple, Slice
        >>> Tuple(0, slice(0, 1)).isempty()
        False
        >>> Tuple(0, slice(0, 0)).isempty()
        True
        >>> Slice(5, 10).isempty()
        False
        >>> Slice(5, 10).isempty(4)
        True

        See Also
        ========
        ndindex.Slice.__len__

        """
        raise NotImplementedError

    def broadcast_arrays(self):
        """
        Broadcast all the array indices in self to a common shape and convert
        boolean array indices into integer array indices.

        The resulting index is equivalent in all contexts where the original
        index is allowed. However, it is possible for the original index to
        give an IndexError but for the new index to not, since integer array
        indices have less stringent shape requirements than boolean array
        indices. There are also some instances for empty indices
        (:any:`isempty` is True) where bounds would be checked before
        broadcasting but not after.

        Any :any:`BooleanArray` indices are converted to :any:`IntegerArray`
        indices. Furthermore, if there are :any:`BooleanArray` or
        :any:`IntegerArray` indices, then any :any:`Integer` indices are also
        converted into scalar :any:`IntegerArray` indices and broadcast.
        Furthermore, if there are multiple boolean scalar indices (`True` or
        `False`), they are combined into a single one.

        Note that array broadcastability is checked in the :any:`Tuple`
        constructor, so this method will not raise any exceptions.

        This is part of what is performed by :any:`expand`, but unlike
        :any:`expand`, this method does not do any other manipulations, and it
        does not require a shape.

        >>> from ndindex import Tuple
        >>> idx = Tuple([[False], [True], [True]], [[4], [5], [5]], -1)
        >>> print(idx.broadcast_arrays())
        Tuple(IntegerArray([[1 2] [1 2] [1 2]]),
              IntegerArray([[0 0] [0 0] [0 0]]),
              IntegerArray([[4 4] [5 5] [5 5]]),
              IntegerArray([[-1 -1] [-1 -1] [-1 -1]]))

        See Also
        ========

        expand

        """
        return self

def asshape(shape, axis=None):
    """
    Cast `shape` as a valid NumPy shape.

    The input can be an integer `n`, which is equivalent to `(n,)`, or a tuple
    of integers.

    If the `axis` argument is provided, an `IndexError` is raised if it is out
    of bounds for the shape.

    The resulting shape is always a tuple of nonnegative integers.

    All ndindex functions that take a shape input should use::

        shape = asshape(shape)

    or::

        shape = asshape(shape, axis=axis)

    """
    from .integer import Integer
    from .tuple import Tuple
    if isinstance(shape, (Tuple, Integer)):
        raise TypeError("ndindex types are not meant to be used as a shape - "
                        "did you mean to use the built-in tuple type?")

    if isinstance(shape, numbers.Number):
        shape = (operator_index(shape),)

    try:
        l = len(shape)
    except TypeError:
        raise TypeError("expected sequence object with len >= 0 or a single integer")

    newshape = []
    # numpy uses __getitem__ rather than __iter__ to index into shape, so we
    # match that
    for i in range(l):
        # Raise TypeError if invalid
        newshape.append(operator_index(shape[i]))

        if shape[i] < 0:
            raise ValueError("unknown (negative) dimensions are not supported")

    if axis is not None:
        if len(newshape) <= axis:
            raise IndexError(f"too many indices for array: array is {len(shape)}-dimensional, but {axis + 1} were indexed")

    return tuple(newshape)


def operator_index(idx):
    """
    Convert `idx` into an integer index using `__index__()` or raise
    `TypeError`.

    This is the same as `operator.index()` except it disallows boolean types.

    This is a slight break in NumPy compatibility, as NumPy allows bools in
    some contexts where `__index__()` is used, for instance, in slices. It
    does disallow it in others, such as in shapes. The main motivation for
    disallowing bools entirely is 1) `numpy.bool_.__index__()` is deprecated
    (currently it matches the built-in `bool.__index__()` and returns the
    object unchanged, but prints a deprecation warning), and 2) for raw
    indices, booleans and `0`/`1` are completely different, i.e., `a[True]` is
    *not* the same as `a[1]`.

    >>> from ndindex.ndindex import operator_index
    >>> operator_index(1)
    1
    >>> operator_index(1.0)
    Traceback (most recent call last):
    ...
    TypeError: 'float' object cannot be interpreted as an integer
    >>> operator_index(True)
    Traceback (most recent call last):
    ...
    TypeError: 'bool' object cannot be interpreted as an integer

    """
    import operator

    if isinstance(idx, bool):
        raise TypeError("'bool' object cannot be interpreted as an integer")
    if isinstance(idx, bool_):
        raise TypeError("'np.bool_' object cannot be interpreted as an integer")
    return operator.index(idx)



from pytest import raises

from numpy import array, arange, isin, prod, unique, intp

from hypothesis import given, assume, example
from hypothesis.strategies import integers, one_of

from ..ndindex import ndindex
from ..integerarray import IntegerArray
from ..tuple import Tuple
from .helpers import ndindices, short_shapes, assert_equal


@example((slice(0, 8), slice(0, 9), slice(0, 10)),
         ([2, 5, 6, 7], slice(1, 9, 1), slice(5, 10, 1)),
         (20, 20, 20))
@example((), (None, array([], dtype=intp)), 0)
@example((), array([], dtype=bool), 0)
@example((), IntegerArray(0), 1)
@example((), (None, True), 0)
@example(..., False, 0)
@example(..., (False,), 0)
@example((IntegerArray(2),), (slice(0, 1),), 3)
@example(([0],), (slice(1, 2),), 3)
@example(array([True]), slice(0, 1), 2)
@example((slice(0, 2), slice(0, 2)), (array([0, 2]), array([0, 2])), 3)
@example(IntegerArray([0, 1]), (slice(0, 1),), 2)
@example(array([1]), slice(0, 1), 1)
@example(IntegerArray(0), slice(0, 1), 3)
@example(IntegerArray(0), slice(1, 2), 3)
@example(IntegerArray(2), slice(0, 1), 3)
@example(..., (-1, array([0])), (1, 1))
@example(..., (-1, array([0, 0])), (1, 1))
@example((slice(None, 1, None), slice(None, 1, None)),
         (array(0), array([0, 0])),
         (1, 1))
@example([[0, 11], [0, 0]], slice(0, 10), 20)
@example(slice(0, 0), 9007199254741193, 1)
@example((0,), (slice(1, 2),), 3)
@example(slice(0, 10), slice(5, 15), 20)
@example((), (slice(None, None, -1),), (2,))
@example((), (..., slice(None, None, -1),), (2,))
@example((slice(0, 1),), (2,), (3,))
@example((slice(0, 5), slice(0, 5)), (slice(3, 10), slice(3, 10)), (20, 20))
@example((slice(0, 5), slice(0, 5)), (1, 1), (10, 10))
@example(0, slice(0, 0), 1)
@example([0], slice(0, 0), 1)
@example(0, slice(0, 1), 1)
@example([0], slice(0, 1), 1)
@example(slice(0, 5), 2, 10)
@example(0, (slice(None, 0, None), Ellipsis), 1)
@example(0, (slice(1, 2),), 1)
@given(ndindices, ndindices, one_of(integers(0, 100), short_shapes))
def test_as_subindex_hypothesis(idx1, idx2, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    try:
        index1 = ndindex(idx1)
        index2 = ndindex(idx2)
    except ValueError: # pragma: no cover
        assume(False)

    empty = False
    try:
        Subindex = index1.as_subindex(index2)
    except NotImplementedError:
        return
    except ValueError as e:
        assert "do not intersect" in e.args[0]
        empty = True

    try:
        a1 = a[index1.raw]
        a2 = a[index2.raw]
    except IndexError: # pragma: no cover
        assume(False)
    except DeprecationWarning as w: # pragma: no cover
        if "Out of bound index found. This was previously ignored when the indexing result contained no elements. In the future the index error will be raised. This error occurs either due to an empty slice, or if an array has zero elements even before indexing." in w.args[0]:
            assume(False)
        else: # pragma: no cover
            raise

    if empty:
        assert not isin(a1, a2).any()
        assert not isin(a2, a1).any()
        with raises(ValueError, match="do not intersect"):
            try:
                index2.as_subindex(index1)
            except NotImplementedError:
                raise ValueError('do not intersect')
    else:
        asubindex = a2[Subindex.raw]

        if (isinstance(index2, IntegerArray)
            or (isinstance(index2, Tuple)
                and any(isinstance(i, IntegerArray) for i in index2.args))):
            # isin(x, y) has the same shape as x. If idx2 has an integer array
            # it may index the same element more than once, but idx1 will not.
            assert_equal(unique(asubindex.flatten()), unique(a1[isin(a1, a2)]))
        else:
            assert_equal(asubindex.flatten(), a1[isin(a1, a2)])

        try:
            subindex2 = index2.as_subindex(index1)
        except NotImplementedError:
            return
        asubindex2 = a1[subindex2.raw]
        assert_equal(asubindex2, asubindex)



import inspect

import numpy as np

from hypothesis import given, example, settings

from pytest import raises, warns

from ..ndindex import ndindex, asshape
from ..booleanarray import BooleanArray
from ..integer import Integer
from ..ellipsis import ellipsis
from ..integerarray import IntegerArray
from ..tuple import Tuple
from .helpers import ndindices, check_same, assert_equal

@given(ndindices)
def test_eq(idx):
    index = ndindex(idx)
    new = type(index)(*index.args)

    if isinstance(new.raw, np.ndarray):
        # trying to get a single value out of comparing two arrays requires all sorts of special handling, just let numpy do it
        assert np.array_equal(new.raw, index.raw)
    else:
        assert new.raw == index.raw

    assert (new == index)
    assert (new.raw == index)
    assert (new == index.raw)
    assert (index == new)
    assert (index.raw == new)
    assert (index == new.raw)

    assert (index.raw == index)
    assert hash(new) == hash(index)
    assert (index == index.raw)
    assert not (index == 'a')
    assert not ('a' == index)
    assert (index != 'a')
    assert ('a' != index)

    try:
        h = hash(idx)
    except TypeError:
        pass
    else:
        assert hash(index) == h

    try:
        h = hash(index.raw)
    except TypeError:
        pass
    else:
        assert hash(index) == h

def test_eq_array_raises():
    index = ndindex([1, 2, 3])
    with raises(TypeError):
        np.equal(index.raw, index)
    with raises(TypeError):
        np.array_equal(index.raw, index)

def test_eq_explicit():
    assert Integer(0) != False
    assert Integer(1) != True
    assert Integer(0) != IntegerArray(0)
    assert IntegerArray([0, 1]) != [False, True]
    assert IntegerArray([0, 1]) == [0, 1]
    assert BooleanArray([False, True]) != [0, 1]
    assert BooleanArray([False, True]) == [False, True]

@example((np.array([1, 2]), 0))
@example([1, 2, 3])
@given(ndindices)
def test_ndindex(idx):
    index = ndindex(idx)
    assert index == idx
    def test_raw_eq(idx, index):
        if isinstance(idx, np.ndarray):
            assert_equal(index.raw, idx)
        elif isinstance(idx, list):
            assert index.dtype in [np.intp, np.bool_]
            assert_equal(index.raw, np.asarray(idx, dtype=index.dtype))
        elif isinstance(idx, tuple):
            assert type(index.raw) == type(idx)
            assert len(index.raw) == len(idx)
            assert index.args == index.raw
            for i, j in zip(idx, index.args):
                test_raw_eq(i, j)
        else:
            assert index.raw == idx
    test_raw_eq(idx, index)
    assert ndindex(index.raw) == index

def test_ndindex_invalid():
    a = np.arange(10)
    for idx in [1.0, [1.0], np.array([1.0]), np.array([1], dtype=object),
                np.array([])]:
        check_same(a, idx)

    # This index is allowed by NumPy, but gives a deprecation warnings. We are
    # not going to allow indices that give deprecation warnings in ndindex.
    with warns(None) as r: # Make sure no warnings are emitted from ndindex()
        raises(IndexError, lambda: ndindex([1, []]))
    assert not r

def test_ndindex_ellipsis():
    raises(IndexError, lambda: ndindex(ellipsis))

def test_signature():
    sig = inspect.signature(Integer)
    assert sig.parameters.keys() == {'idx'}


@example(([0, 1],))
@example((IntegerArray([], (0, 1)),))
@example(IntegerArray([], (0, 1)))
@example((1, ..., slice(1, 2)))
# eval can sometimes be slower than the default deadline of 200ms for large
# array indices
@settings(deadline=None)
@given(ndindices)
def test_repr_str(idx):
    # The repr form should be re-creatable
    index = ndindex(idx)
    d = {}
    exec("from ndindex import *", d)
    assert eval(repr(index), d) == idx

    # Str may not be re-creatable. Just test that it doesn't give an exception.
    str(index)

def test_asshape():
    assert asshape(1) == (1,)
    assert asshape(np.int64(2)) == (2,)
    assert type(asshape(np.int64(2))[0]) == int
    assert asshape((1, 2)) == (1, 2)
    assert asshape([1, 2]) == (1, 2)
    assert asshape((np.int64(1), np.int64(2))) == (1, 2)
    assert type(asshape((np.int64(1), np.int64(2)))[0]) == int
    assert type(asshape((np.int64(1), np.int64(2)))[1]) == int

    raises(TypeError, lambda: asshape(1.0))
    raises(TypeError, lambda: asshape((1.0,)))
    raises(ValueError, lambda: asshape(-1))
    raises(ValueError, lambda: asshape((1, -1)))
    raises(TypeError, lambda: asshape(...))
    raises(TypeError, lambda: asshape(Integer(1)))
    raises(TypeError, lambda: asshape(Tuple(1, 2)))
    raises(TypeError, lambda: asshape((True,)))



from collections.abc import Sequence
from itertools import product
from functools import reduce
from operator import mul

from .ndindex import ImmutableObject, operator_index, asshape, ndindex
from .tuple import Tuple
from .slice import Slice
from .integer import Integer
from .subindex_helpers import ceiling

# np.prod has overflow and math.prod is Python 3.8+ only
def prod(seq):
    return reduce(mul, seq, 1)

class ChunkSize(ImmutableObject, Sequence):
    """
    Represents a chunk size tuple.

    A chunk size is a tuple of length n where each element is either a
    positive integer or `None`. It represents a chunking of an array with n
    dimensions, where each corresponding dimension is chunked by the
    corresponding chunk size, or not chunked for `None` (note, `None` chunks
    are currently not yet implemented).

    For example, given a 3 dimensional chunk size of `(20, 20, None)` and an
    array of shape `(40, 30, 10)`, the array would be split into four chunks,
    corresponding to the indices `0:20,0:20,:`, `0:20,20:30,:`,
    `20:40,0:20,:`, and `20:40,20:30,:`. Note that the size of a chunk may be
    less than the total chunk size if the array shape is not a multiple of the
    chunk size in a given dimension.

    `ChunkSize` behaves like a `tuple`. For example, `chunk_size[0]` gives the
    first chunk dimension, and `len(chunk_size)` gives the number of
    dimensions of a chunk. Also, the input to ChunkSize should be a tuple,
    just as with the `tuple` constructor, even for single dimensional chunk
    sizes.

    >>> from ndindex import ChunkSize
    >>> ChunkSize((20, 30, 40))
    ChunkSize((20, 30, 40))
    >>> ChunkSize((2**12,))
    ChunkSize((4096,))

    """
    def _typecheck(self, chunk_size):
        # TODO: Also accept ChunkSize(1, 2, 3)?
        if isinstance(chunk_size, Tuple):
            raise TypeError("Tuple is not a valid input to ChunkSize. Use tuple instead.")
        args = []
        for i in chunk_size:
            if i is None:
                raise NotImplementedError("None in chunks is not supported yet")
                # args.append(i)
            else:
                try:
                    i = operator_index(i)
                except TypeError:
                    raise TypeError("Chunks must be positive integers or None")
                if i <= 0:
                    raise ValueError("Chunks must be positive integers")
                args.append(i)
        return (tuple(args),)

    def __hash__(self):
        return hash(self.args[0])

    # Methods for collections.abc.Sequence to make ChunkSize act like a tuple
    def __getitem__(self, *args):
        return self.args[0].__getitem__(*args)

    def __len__(self):
        return len(self.args[0])

    def num_chunks(self, shape):
        """
        Give the number of chunks for the given `shape`.

        This is the same as `len(list(self.indices(shape)))`, but much faster.
        `shape` must have the same number of dimensions as `self`.

        >>> from ndindex import ChunkSize
        >>> chunk_size = ChunkSize((10, 10, 10))
        >>> shape = (10000, 10000, 10000)
        >>> # len(list(chunk_size.indices(shape))) would be very slow, as
        >>> # would have to iterate all 1 billion chunks
        >>> chunk_size.num_chunks(shape)
        1000000000

        See Also
        ========

        ChunkSize.indices

        """
        shape = asshape(shape)
        d = [ceiling(i, c) for i, c in zip(shape, self)]
        if 0 in d:
            return 1
        return prod(d)

    def indices(self, shape):
        """
        Yield a set of ndindex indices for the chunks on an array of shape `shape`.

        `shape` should have the same number of dimensions as `self`. If the
        shape is not a multiple of the chunk size, some chunks will be
        truncated, so that :any:`len(idx.args[i]) <ndindex.Slice.__len__>` can
        be used to get the size of an indexed axis.

        For example, if `a` has shape `(10, 19)` and is chunked into chunks
        of shape `(5, 5)`:

        >>> from ndindex import ChunkSize
        >>> chunk_size = ChunkSize((5, 5))
        >>> for idx in chunk_size.indices((10, 19)):
        ...     print(idx)
        Tuple(slice(0, 5, 1), slice(0, 5, 1))
        Tuple(slice(0, 5, 1), slice(5, 10, 1))
        Tuple(slice(0, 5, 1), slice(10, 15, 1))
        Tuple(slice(0, 5, 1), slice(15, 19, 1))
        Tuple(slice(5, 10, 1), slice(0, 5, 1))
        Tuple(slice(5, 10, 1), slice(5, 10, 1))
        Tuple(slice(5, 10, 1), slice(10, 15, 1))
        Tuple(slice(5, 10, 1), slice(15, 19, 1))

        See Also
        ========

        ChunkSize.num_chunks

        """
        shape = asshape(shape)

        if len(shape) != len(self):
            raise ValueError("chunks dimensions must equal the array dimensions")
        d = [ceiling(i, c) for i, c in zip(shape, self)]
        if 0 in d:
            yield Tuple(*[Slice(0, bool(i)*chunk_size, 1) for i, chunk_size in zip(d, self)]).expand(shape)
        for p in product(*[range(i) for i in d]):
            # p = (0, 0, 0), (0, 0, 1), ...
            yield Tuple(*[Slice(chunk_size*i, min(chunk_size*(i + 1), n), 1)
                          for n, chunk_size, i in zip(shape, self, p)])

    def as_subchunks(self, idx, shape, *, _force_slow=False):
        """
        Split an index `idx` on an array of shape `shape` into subchunk indices.

        Yields indices `c`, where `c` is an index for the chunk that
        should be sliced. Only those `c` for which `idx` includes at least one
        element are yielded.

        That is to say, for each `c` index yielded, `a[c][idx.as_subindex(c)]`
        will give those elements of `a[idx]` that are part of the `c` chunk,
        and together they give all the elements of `a[idx]`. See also the
        docstring of :meth:`~ndindex.NDIndex.as_subindex`.

        This method is roughly equivalent to

        .. code:: python

           def as_subchunks(self, idx, shape):
               for c in self.indices(shape):
                   try:
                       index = idx.as_subindex(c)
                   except ValueError:
                       # as_subindex raises ValueError in some cases when the
                       # indices do not intersect (see the docstring of
                       # as_subindex())
                       continue

                   if not index.isempty(self):
                       # Yield those c for which idx.as_subindex(c) is nonempty
                       yield c

        except it is more efficient.

        >>> from ndindex import ChunkSize, Tuple
        >>> idx = Tuple(slice(5, 15), 0)
        >>> shape = (20, 20)
        >>> chunk_size = ChunkSize((10, 10))
        >>> for c in chunk_size.as_subchunks(idx, shape):
        ...     print(c)
        ...     print('    ', idx.as_subindex(c))
        Tuple(slice(0, 10, 1), slice(0, 10, 1))
            Tuple(slice(5, 10, 1), 0)
        Tuple(slice(10, 20, 1), slice(0, 10, 1))
            Tuple(slice(0, 5, 1), 0)

        See Also
        ========

        ndindex.NDIndex.as_subindex

        """
        shape = asshape(shape)
        if len(shape) != len(self):
            raise ValueError("chunks dimensions must equal the array dimensions")

        if 0 in shape:
            return
        idx = ndindex(idx).expand(shape)

        # The slow naive fallback is kept here for testing purposes and to support
        # indices that aren't supported in the fast way yet below.
        def _fallback():
            for c in self.indices(shape):
                try:
                    index = idx.as_subindex(c)
                except ValueError:
                    continue

                if not index.isempty(self):
                    yield c
            return

        if _force_slow or len(idx.args) > len(self):
            yield from _fallback()
            return

        iters = []
        for i, n in zip(idx.args, self):
            if isinstance(i, Integer):
                iters.append([i.raw//n])
            elif isinstance(i, Slice) and i.step > 0:
                a, N, m = i.args
                if m > n:
                    iters.append([(a + k*m)//n for k in range(ceiling(N-a, m))])
                else:
                    iters.append(range(a//n, ceiling(N, n)))
            else:
                # fallback to the naive algorithm
                yield from _fallback()
                return

        def _indices(iters):
            for p in product(*iters):
                # p = (0, 0, 0), (0, 0, 1), ...
                yield Tuple(*[Slice(chunk_size*i, min(chunk_size*(i + 1), n), 1)
                          for n, chunk_size, i in zip(shape, self, p)])

        for c in _indices(iters):
            # Empty indices should be impossible by the construction of the
            # iterators above.
            yield c



"""
This file has the main algorithm for Slice.as_subindex(Slice)

Since Integer can use the same algorithm via Slice(i, i+1), and IntegerArray
needs to do this but in a way that only uses array friendly operations, we
need to have this factored out into a separately callable function.

TODO: we could remove the dependency on SymPy if we wanted to, by implementing
the special cases for ilcm(a, b) and the Chinese Remainder Theorem for 2
equations. It wouldn't be too bad (it just requires the extended gcd
algorithm), but depending on SymPy also isn't a big deal for the time being.

"""
from numpy import broadcast_arrays, amin, amax, where

def _crt(m1, m2, v1, v2):
    """
    Chinese Remainder Theorem

    Returns x such that x = v1 (mod m1) and x = v2 (mod m2), or None if no
    such solution exists.

    """
    # Avoid calling sympy_crt in the cases where the inputs would be arrays.
    if m1 == 1:
        return v2 % m2
    if m2 == 1:
        return v1 % m1

    # Only import SymPy when necessary
    from sympy.ntheory.modular import crt as sympy_crt

    res = sympy_crt([m1, m2], [v1, v2])
    if res is None:
        return res
    # Make sure the result isn't a gmpy2.mpz
    return int(res[0])

def _ilcm(a, b):
    # Avoid calling sympy_ilcm in the cases where the inputs would be arrays.
    if a == 1:
        return b
    if b == 1:
        return a

    # Only import SymPy when necessary
    from sympy import ilcm as sympy_ilcm

    return sympy_ilcm(a, b)

def ceiling(a, b):
    """
    Returns ceil(a/b)
    """
    return -(-a//b)

def _max(a, b):
    if isinstance(a, int) and isinstance(b, int):
        return max(a, b)
    return amax(broadcast_arrays(a, b), axis=0)

def _min(a, b):
    if isinstance(a, int) and isinstance(b, int):
        return min(a, b)
    return amin(broadcast_arrays(a, b), axis=0)

def _smallest(x, a, m):
    """
    Gives the smallest integer >= x that equals a (mod m)

    Assumes x >= 0, m >= 1, and 0 <= a < m.
    """
    n = ceiling(x - a, m)
    return a + n*m

def subindex_slice(s_start, s_stop, s_step, i_start, i_stop, i_step):
    """
    Computes s.as_subindex(i) for slices s and i in a way that is (mostly)
    compatible with NumPy arrays.

    Returns (start, stop, step).

    """
    # Chinese Remainder Theorem. We are looking for a solution to
    #
    # x = s.start (mod s.step)
    # x = index.start (mod index.step)
    #
    # If crt() returns None, then there are no solutions (the slices do
    # not overlap).
    common = _crt(s_step, i_step, s_start, i_start)

    if common is None:
        return (0, 0, 1)
    lcm = _ilcm(s_step, i_step)
    start = _max(s_start, i_start)

    # Get the smallest lcm multiple of common that is >= start
    start = _smallest(start, common, lcm)
    # Finally, we need to shift start so that it is relative to index
    start = (start - i_start)//i_step

    stop = ceiling((_min(s_stop, i_stop) - i_start), i_step)
    stop = where(stop < 0, 0, stop)

    step = lcm//i_step # = s_step//igcd(s_step, i_step)

    return (start, stop, step)



from numpy import (broadcast, broadcast_to, array, intp, ndarray, bool_,
                   logical_and, broadcast_arrays)

from .ndindex import NDIndex, ndindex, asshape
from .subindex_helpers import subindex_slice

class Tuple(NDIndex):
    """
    Represents a tuple of single-axis indices.

    Valid single axis indices are

    - :class:`Integer`
    - :class:`Slice`
    - :class:`ellipsis`
    - :class:`Newaxis`
    - :class:`IntegerArray`
    - :class:`BooleanArray`

    (some of the above are not yet implemented)

    `Tuple(x1, x2, …, xn)` represents the index `a[x1, x2, …, xn]` or,
    equivalently, `a[(x1, x2, …, xn)]`. `Tuple()` with no arguments is the
    empty tuple index, `a[()]`, which returns `a` unchanged.

    >>> from ndindex import Tuple, Slice
    >>> import numpy as np
    >>> idx = Tuple(0, Slice(2, 4))
    >>> a = np.arange(10).reshape((2, 5))
    >>> a
    array([[0, 1, 2, 3, 4],
           [5, 6, 7, 8, 9]])
    >>> a[0, 2:4]
    array([2, 3])
    >>> a[idx.raw]
    array([2, 3])

    .. note::

       `Tuple` does *not* represent a tuple, but rather an *tuple index*. It
       does not have most methods that `tuple` has, and should not be used in
       non-indexing contexts. See the document on :any:`type-confusion` for
       more details.

    """
    def _typecheck(self, *args):
        from .array import ArrayIndex
        from .ellipsis import ellipsis
        from .newaxis import Newaxis
        from .slice import Slice
        from .integer import Integer
        from .booleanarray import BooleanArray, _is_boolean_scalar

        newargs = []
        arrays = []
        array_block_start = False
        array_block_stop = False
        has_array = any(isinstance(i, (ArrayIndex, list, ndarray, bool, bool_)) for i in args)
        has_boolean_scalar = False
        for arg in args:
            newarg = ndindex(arg)
            if isinstance(newarg, Tuple):
                if len(args) == 1:
                    raise ValueError("tuples inside of tuple indices are not supported. Did you mean to call Tuple(*args) instead of Tuple(args)?")
                raise ValueError("tuples inside of tuple indices are not supported. If you meant to use a fancy index, use a list or array instead.")
            newargs.append(newarg)
            if isinstance(newarg, ArrayIndex):
                array_block_start = True
                if _is_boolean_scalar(newarg):
                    has_boolean_scalar = True
                elif isinstance(newarg, BooleanArray):
                    arrays.extend(newarg.raw.nonzero())
                else:
                    arrays.append(newarg.raw)
            elif has_array and isinstance(newarg, Integer):
                array_block_start = True
            if isinstance(newarg, (Slice, ellipsis, Newaxis)) and array_block_start:
                array_block_stop = True
            elif isinstance(newarg, (ArrayIndex, Integer)):
                if array_block_start and array_block_stop:
                    # If the arrays in a tuple index are separated by a slice,
                    # ellipsis, or newaxis, the behavior is that the
                    # dimensions indexed by the array (and integer) indices
                    # are added to the front of the final array shape. Travis
                    # told me that he regrets implementing this behavior in
                    # NumPy and that he wishes it were in error. So for now,
                    # that is what we are going to do, unless it turns out
                    # that we actually need it.
                    raise NotImplementedError("Array indices separated by slices, ellipses (...), or newaxes (None) are not supported")

        if newargs.count(...) > 1:
            raise IndexError("an index can only have a single ellipsis ('...')")
        if len(arrays) > 0:
            if has_boolean_scalar:
                raise NotImplementedError("Tuples mixing boolean scalars (True or False) with arrays are not yet supported.")

            try:
                broadcast(*[i for i in arrays])
            except ValueError as e:
                assert e.args == ("shape mismatch: objects cannot be broadcast to a single shape",)
                raise IndexError("shape mismatch: indexing arrays could not be broadcast together with shapes %s" % ' '.join([str(i.shape) for i in arrays]))

        return tuple(newargs)

    def __eq__(self, other):
        if isinstance(other, tuple):
            return self.args == other
        elif isinstance(other, Tuple):
            return self.args == other.args
        return False

    def __hash__(self):
        # Since self.args is itself a tuple, it will match the hash of
        # self.raw when it is hashable.
        return hash(self.args)

    def __repr__(self):
        from .array import ArrayIndex
        # Since tuples are nested, we can print the raw form of the args to
        # make them a little more readable.
        def _repr(s):
            if s == ...:
                return '...'
            if isinstance(s, ArrayIndex):
                if s.shape and 0 not in s.shape:
                    return repr(s.array.tolist())
                return repr(s)
            return repr(s.raw)
        return f"{self.__class__.__name__}({', '.join(map(_repr, self.args))})"

    def __str__(self):
        from .array import ArrayIndex
        # Since tuples are nested, we can print the raw form of the args to
        # make them a little more readable.
        def _str(s):
            if s == ...:
                return '...'
            if isinstance(s, ArrayIndex):
                return str(s)
            return str(s.raw)
        return f"{self.__class__.__name__}({', '.join(map(_str, self.args))})"

    @property
    def has_ellipsis(self):
        """
        Returns True if self has an ellipsis
        """
        return ... in self.args

    @property
    def ellipsis_index(self):
        """
        Give the index i of `self.args` where the ellipsis is.

        If `self` doesn't have an ellipsis, it gives `len(self.args)`, since
        tuple indices without an ellipsis always implicitly end in an
        ellipsis.

        The resulting value `i` is such that `self.args[:i]` indexes the
        beginning axes of an array and `self.args[i+1:]` indexes the end axes
        of an array.

        >>> from ndindex import Tuple
        >>> idx = Tuple(0, 1, ..., 2, 3)
        >>> i = idx.ellipsis_index
        >>> i
        2
        >>> idx.args[:i]
        (Integer(0), Integer(1))
        >>> idx.args[i+1:]
        (Integer(2), Integer(3))

        >>> Tuple(0, 1).ellipsis_index
        2

        """
        if self.has_ellipsis:
            return self.args.index(...)
        return len(self.args)

    @property
    def raw(self):
        return tuple(i.raw for i in self.args)

    def reduce(self, shape=None):
        r"""
        Reduce a Tuple index on an array of shape `shape`

        A `Tuple` with a single argument is always reduced to that single
        argument (because `a[idx,]` is the same as `a[idx]`).

        >>> from ndindex import Tuple

        >>> Tuple(slice(2, 4)).reduce()
        Slice(2, 4, 1)

        If an explicit array shape is given, the result will either be
        `IndexError` if the index is invalid for the given shape, or an index
        that is as simple as possible:

        - All the elements of the :any:`Tuple` are recursively :any:`reduced
          <NDIndex.reduce>`.

        - Any axes that can be merged into an :any:`ellipsis` are removed.
          This includes the implicit ellipsis at the end of a Tuple that
          doesn't contain any explicit ellipses.

        - :any:`Ellipses <ellipsis>` that don't match any axes are removed.

        - An :any:`ellipsis` at the end of the :any:`Tuple` is removed.

        - Scalar :any:`BooleanArray` arguments (`True` or `False`) are
          combined into a single term (the first boolean scalar is replaced
          with the AND of all the boolean scalars).

        - If the resulting :any:`Tuple` would have a single argument, that
          argument is returned.

        >>> idx = Tuple(0, ..., slice(0, 3))
        >>> idx.reduce((5, 4))
        Tuple(0, slice(0, 3, 1))
        >>> idx.reduce((5, 3))
        Integer(0)

        >>> idx = Tuple(slice(0, 10), -3)
        >>> idx.reduce((5,))
        Traceback (most recent call last):
        ...
        IndexError: too many indices for array: array is 1-dimensional, but 2 were indexed
        >>> idx.reduce((5, 2))
        Traceback (most recent call last):
        ...
        IndexError: index -3 is out of bounds for axis 1 with size 2

        Note
        ====

        ndindex presently does not distinguish between scalar objects and
        rank-0 arrays. It is possible for the original index to produce one
        and the reduced index to produce the other. In particular, the
        presence of a redundant ellipsis forces NumPy to return a rank-0 array
        instead of a scalar.

        >>> import numpy as np
        >>> a = np.array([0, 1])
        >>> Tuple(..., 1).reduce(a.shape)
        Integer(1)
        >>> a[..., 1]
        array(1)
        >>> a[1]
        1

        See https://github.com/Quansight-Labs/ndindex/issues/22.

        See Also
        ========

        .Tuple.expand
        .NDIndex.reduce
        .Slice.reduce
        .Integer.reduce
        .ellipsis.reduce
        .Newaxis.reduce
        .IntegerArray.reduce
        .BooleanArray.reduce

        """
        from .slice import Slice
        from .integer import Integer
        from .booleanarray import BooleanArray, _is_boolean_scalar
        from .integerarray import IntegerArray

        args = list(self.args)
        if ... not in args:
            return type(self)(*args, ...).reduce(shape)

        boolean_scalars = [i for i in args if _is_boolean_scalar(i)]
        if len(boolean_scalars) > 1:
            _args = []
            seen_boolean_scalar = False
            for s in args:
                if _is_boolean_scalar(s):
                    if seen_boolean_scalar:
                        continue
                    _args.append(BooleanArray(all(i == True for i in boolean_scalars)))
                    seen_boolean_scalar = True
                else:
                    _args.append(s)
            return type(self)(*_args).reduce(shape)

        arrays = []
        for i in args:
            if _is_boolean_scalar(i):
                continue
            elif isinstance(i, IntegerArray):
                arrays.append(i.raw)
            elif isinstance(i, BooleanArray):
                # TODO: Avoid explicitly calling nonzero
                arrays.extend(i.raw.nonzero())
        if not arrays:
            # Older versions of NumPy do not allow broadcast() with no arguments
            broadcast_shape = ()
        else:
            broadcast_shape = broadcast(*arrays).shape
        # If the broadcast shape is empty, out of bounds indices in
        # non-empty arrays are ignored, e.g., ([], [10]) would broadcast to
        # ([], []), so the bounds for 10 are not checked. Thus, we must do
        # this before calling reduce() on the arguments. This rule, however,
        # is *not* followed for scalar integer indices.
        if 0 in broadcast_shape:
            for i in range(len(args)):
                s = args[i]
                if isinstance(s, IntegerArray):
                    if s.ndim == 0:
                        args[i] = Integer(s.raw)
                    else:
                        # broadcast_to(x) gives a readonly view on x, which is also
                        # readonly, so set _copy=False to avoid representing the full
                        # broadcasted array in memory.
                        args[i] = type(s)(broadcast_to(s.raw, broadcast_shape),
                                          _copy=False)

        if shape is not None:
            # assert self.args.count(...) == 1
            # assert self.args.count(False) <= 1
            # assert self.args.count(True) <= 1
            n_newaxis = self.args.count(None)
            n_boolean = sum(i.ndim - 1 for i in args if
                            isinstance(i, BooleanArray) and not _is_boolean_scalar(i))
            if True in args or False in args:
                n_boolean -= 1
            indexed_args = len(args) + n_boolean - n_newaxis - 1 # -1 for the

            shape = asshape(shape, axis=indexed_args - 1)

        ellipsis_i = self.ellipsis_index

        preargs = []
        removable = shape is not None
        begin_offset = args[:ellipsis_i].count(None)
        begin_offset -= sum(j.ndim - 1 for j in args[:ellipsis_i] if
                            isinstance(j, BooleanArray))
        for i, s in enumerate(reversed(args[:ellipsis_i]), start=1):
            if s == None:
                begin_offset -= 1
            elif isinstance(s, BooleanArray):
                begin_offset += s.ndim - 1
            axis = ellipsis_i - i - begin_offset
            reduced = s.reduce(shape, axis=axis)
            if (removable
                and isinstance(reduced, Slice)
                and reduced == Slice(0, shape[axis], 1)):
                continue
            else:
                removable = False
                preargs.insert(0, reduced)

        if shape is None:
            endargs = [s.reduce() for s in args[ellipsis_i+1:]]
        else:
            endargs = []
            end_offset = 0
            for i, s in enumerate(reversed(args[ellipsis_i+1:]), start=1):
                if isinstance(s, BooleanArray):
                    end_offset -= s.ndim - 1
                elif s == None:
                    end_offset += 1
                axis = len(shape) - i + end_offset
                if not (isinstance(s, IntegerArray) and (0 in broadcast_shape or
                                                         False in args)):
                    # Array bounds are not checked when the broadcast shape is empty
                    s = s.reduce(shape, axis=axis)
                endargs.insert(0, s)

        if shape is not None:
            # Remove redundant slices
            axis = len(shape) - len(endargs) + end_offset
            for i, s in enumerate(endargs):
                axis += i
                if (isinstance(s, Slice)
                    and s == Slice(0, shape[axis], 1)):
                    i += 1
                    continue
                else:
                    break
            if endargs:
                endargs = endargs[i:]

        if shape is None or (endargs and len(preargs) + len(endargs)
                             < len(shape) + args.count(None) - n_boolean):
            preargs = preargs + [...]

        newargs = preargs + endargs

        if newargs and newargs[-1] == ...:
            newargs = newargs[:-1]

        if len(newargs) == 1:
            return newargs[0]

        return type(self)(*newargs)

    def broadcast_arrays(self):
        from .booleanarray import BooleanArray, _is_boolean_scalar
        from .integerarray import IntegerArray
        from .integer import Integer

        args = self.args
        boolean_scalars = [i for i in args if _is_boolean_scalar(i)]
        if len(boolean_scalars) > 1:
            _args = []
            seen_boolean_scalar = False
            for s in args:
                if _is_boolean_scalar(s):
                    if seen_boolean_scalar:
                        continue
                    _args.append(BooleanArray(all(i == True for i in boolean_scalars)))
                    seen_boolean_scalar = True
                else:
                    _args.append(s)
            return type(self)(*_args).broadcast_arrays()

        # Broadcast all array indices. Note that broadcastability is checked
        # in the Tuple constructor, so this should not fail.
        boolean_nonzero = {}
        arrays = []
        for s in args:
            if _is_boolean_scalar(s):
                continue
            elif isinstance(s, IntegerArray):
                arrays.append(s.raw)
            elif isinstance(s, BooleanArray):
                nz = s.raw.nonzero()
                arrays.extend(nz)
                boolean_nonzero[s] = nz
        if not arrays:
            return self
        broadcast_shape = broadcast(*arrays).shape

        newargs = []
        for s in args:
            if isinstance(s, BooleanArray):
                if not _is_boolean_scalar(s):
                    newargs.extend([IntegerArray(broadcast_to(i, broadcast_shape))
                                    for i in boolean_nonzero[s]])
            elif isinstance(s, Integer):
                # broadcast_to(x) gives a readonly view on x, which is also
                # readonly, so set _copy=False to avoid representing the full
                # broadcasted array in memory.
                newargs.append(IntegerArray(broadcast_to(array(s.raw, dtype=intp),
                                              broadcast_shape), _copy=False))
            elif isinstance(s, IntegerArray):
                newargs.append(IntegerArray(broadcast_to(s.raw, broadcast_shape),
                                            _copy=False))
            else:
                newargs.append(s)
        return Tuple(*newargs)

    def expand(self, shape):
        # The expand() docstring is on NDIndex.expand()
        from .array import ArrayIndex
        from .booleanarray import BooleanArray, _is_boolean_scalar
        from .integer import Integer
        from .integerarray import IntegerArray
        from .slice import Slice

        args = list(self.args)
        if ... not in args:
            return type(self)(*args, ...).expand(shape)

        # TODO: Use broadcast_arrays here. The challenge is that we still need
        # to do bounds checks on nonscalar integer arrays that get broadcast
        # away.
        boolean_scalars = [i for i in args if _is_boolean_scalar(i)]
        if len(boolean_scalars) > 1:
            _args = []
            seen_boolean_scalar = False
            for s in args:
                if _is_boolean_scalar(s):
                    if seen_boolean_scalar:
                        continue
                    _args.append(BooleanArray(all(i == True for i in boolean_scalars)))
                    seen_boolean_scalar = True
                else:
                    _args.append(s)
            return type(self)(*_args).expand(shape)

        # Broadcast all array indices. Note that broadcastability is checked
        # in the Tuple constructor, so this should not fail.
        arrays = []
        for i in args:
            if _is_boolean_scalar(i):
                continue
            elif isinstance(i, IntegerArray):
                arrays.append(i.raw)
            elif isinstance(i, BooleanArray):
                # TODO: Avoid calling nonzero twice
                arrays.extend(i.raw.nonzero())
        if not arrays:
            # Older versions of NumPy do not allow broadcast() with no arguments
            broadcast_shape = ()
        else:
            broadcast_shape = broadcast(*arrays).shape
        # If the broadcast shape is empty, out of bounds indices in
        # non-empty arrays are ignored, e.g., ([], [10]) would broadcast to
        # ([], []), so the bounds for 10 are not checked. Thus, we must do
        # this before calling reduce() on the arguments. This rule, however,
        # is *not* followed for scalar integer indices.
        if arrays:
            for i in range(len(args)):
                s = args[i]
                if isinstance(s, IntegerArray):
                    if s.ndim == 0:
                        args[i] = Integer(s.raw)
                    else:
                        # broadcast_to(x) gives a readonly view on x, which is also
                        # readonly, so set _copy=False to avoid representing the full
                        # broadcasted array in memory.
                        args[i] = type(s)(broadcast_to(s.raw, broadcast_shape),
                                          _copy=False)

        # assert args.count(...) == 1
        # assert args.count(False) <= 1
        # assert args.count(True) <= 1
        n_newaxis = args.count(None)
        n_boolean = sum(i.ndim - 1 for i in args if
                        isinstance(i, BooleanArray) and not _is_boolean_scalar(i))
        if True in args or False in args:
            n_boolean -= 1
        indexed_args = len(args) + n_boolean - n_newaxis - 1 # -1 for the ellipsis
        shape = asshape(shape, axis=indexed_args - 1)

        ellipsis_i = self.ellipsis_index

        startargs = []
        begin_offset = 0
        for i, s in enumerate(args[:ellipsis_i]):
            axis = i + begin_offset
            if not (isinstance(s, IntegerArray) and (0 in broadcast_shape or
                                                     False in args)):
                s = s.reduce(shape, axis=axis)
            if isinstance(s, ArrayIndex):
                if isinstance(s, BooleanArray):
                    begin_offset += s.ndim - 1
                    if not _is_boolean_scalar(s):
                        s = s.reduce(shape, axis=axis)
                        startargs.extend([IntegerArray(broadcast_to(i,
                                                                  broadcast_shape))
                                        for i in s.array.nonzero()])
                        continue
            elif arrays and isinstance(s, Integer):
                s = IntegerArray(broadcast_to(array(s.raw, dtype=intp),
                                              broadcast_shape), _copy=False)
            elif s == None:
                begin_offset -= 1
            startargs.append(s)

        # TODO: Merge this with the above loop
        endargs = []
        end_offset = 0
        for i, s in enumerate(reversed(args[ellipsis_i+1:]), start=1):
            if isinstance(s, ArrayIndex):
                if isinstance(s, BooleanArray):
                    end_offset -= s.ndim - 1
                    if not _is_boolean_scalar(s):
                        s = s.reduce(shape, axis=len(shape) - i + end_offset)
                        endargs.extend([IntegerArray(broadcast_to(i,
                                                                  broadcast_shape))
                                        for i in reversed(s.array.nonzero())])
                        continue
            elif arrays and isinstance(s, Integer):
                if (0 in broadcast_shape or False in args):
                    s = s.reduce(shape, axis=len(shape)-i+end_offset)
                s = IntegerArray(broadcast_to(array(s.raw, dtype=intp),
                                              broadcast_shape), _copy=False)
            elif s == None:
                end_offset += 1
            axis = len(shape) - i + end_offset
            assert axis >= 0
            if not (isinstance(s, IntegerArray) and (0 in broadcast_shape or
                                                     False in args)):
                # Array bounds are not checked when the broadcast shape is empty
                s = s.reduce(shape, axis=axis)
            endargs.append(s)

        idx_offset = begin_offset - end_offset

        midargs = [Slice(None).reduce(shape, axis=i + ellipsis_i + begin_offset) for
                        i in range(len(shape) - len(args) + 1 - idx_offset)]


        newargs = startargs + midargs + endargs[::-1]

        return type(self)(*newargs)

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        from .array import ArrayIndex
        from .booleanarray import BooleanArray

        shape = asshape(shape)

        if self == Tuple():
            return shape

        # This will raise any IndexErrors
        self = self.expand(shape)

        newshape = []
        axis = 0
        arrays = False
        for i, s in enumerate(self.args):
            if s == None:
                newshape.append(1)
                axis -= 1
            # After expand(), there will be at most one boolean scalar
            elif s == True:
                newshape.append(1)
                axis -= 1
            elif s == False:
                newshape.append(0)
                axis -= 1
            elif isinstance(s, ArrayIndex):
                if not arrays:
                    # Multiple arrays are all broadcast together (in expand())
                    # and iterated as one, so we only need to get the shape
                    # for the first array we see. Note that arrays separated
                    # by ellipses, slices, or newaxes affect the shape
                    # differently, but these are currently unsupported (see
                    # the comments in the Tuple constructor).

                    # expand() should remove all non scalar boolean arrays
                    assert not isinstance(s, BooleanArray)

                    newshape.extend(list(s.newshape(shape[axis])))
                    arrays = True
            else:
                newshape.extend(list(s.newshape(shape[axis])))
            axis += 1
        return tuple(newshape)

    def as_subindex(self, index):
        from .ndindex import ndindex
        from .array import ArrayIndex
        from .slice import Slice
        from .integer import Integer
        from .integerarray import IntegerArray
        from .booleanarray import BooleanArray

        index = ndindex(index).reduce().broadcast_arrays()

        self = self.broadcast_arrays()

        if ... in self.args:
            raise NotImplementedError("Tuple.as_subindex() is not yet implemented for tuples with ellipses")

        if isinstance(index, (Integer, ArrayIndex, Slice)):
            index = Tuple(index)
        if isinstance(index, Tuple):
            new_args = []
            boolean_arrays = []
            integer_arrays = []
            if any(isinstance(i, Slice) and i.step < 0 for i in index.args):
                    raise NotImplementedError("Tuple.as_subindex() is only implemented on slices with positive steps")
            if ... in index.args:
                raise NotImplementedError("Tuple.as_subindex() is not yet implemented for tuples with ellipses")
            for self_arg, index_arg in zip(self.args, index.args):
                if (isinstance(self_arg, IntegerArray) and
                    isinstance(index_arg, Slice)):
                    if (self_arg.array < 0).any():
                        raise NotImplementedError("IntegerArray.as_subindex() is only implemented for arrays with all nonnegative entries. Try calling reduce() with a shape first.")
                    if index_arg.step < 0:
                        raise NotImplementedError("IntegerArray.as_subindex(Slice) is only implemented for slices with positive steps")

                    # After reducing, start is not None when step > 0
                    if index_arg.stop is None or index_arg.start < 0 or index_arg.stop < 0:
                        raise NotImplementedError("IntegerArray.as_subindex(Slice) is only implemented for slices with nonnegative start and stop. Try calling reduce() with a shape first.")

                    s = self_arg.array
                    start, stop, step = subindex_slice(
                        s, s+1, 1, index_arg.start, index_arg.stop, index_arg.step)
                    if (stop <= 0).all():
                        raise ValueError("Indices do not intersect")
                    if start.shape == ():
                        if start >= stop:
                            raise ValueError("Indices do not intersect")

                    integer_arrays.append((start, stop))
                    # Placeholder. We need to mask out the stops below.
                    new_args.append(IntegerArray(start))
                else:
                    subindex = self_arg.as_subindex(index_arg)
                    if isinstance(subindex, Tuple):
                        assert subindex == ()
                        subindex # Workaround https://github.com/nedbat/coveragepy/issues/1029
                        continue
                    if isinstance(subindex, BooleanArray):
                        boolean_arrays.append(subindex)
                    new_args.append(subindex)
            args_remainder = self.args[min(len(self.args), len(index.args)):]
            index_remainder = index.args[min(len(self.args), len(index.args)):]
            if any(isinstance(i, ArrayIndex) and i.isempty() for i in
                   index_remainder):
                raise ValueError("Indices do not intersect")
            for arg in args_remainder:
                if isinstance(arg, BooleanArray):
                    boolean_arrays.append(arg)
                if isinstance(arg, IntegerArray):
                    integer_arrays.append((arg.array, arg.array+1))
                new_args.append(arg)
            # Replace all boolean arrays with the logical AND of them.
            if any(i.isempty() for i in boolean_arrays):
                raise ValueError("Indices do not intersect")
            if boolean_arrays:
                if len(boolean_arrays) > 1:
                    new_array = BooleanArray(logical_and.reduce([i.array for i in boolean_arrays]))
                else:
                    new_array = boolean_arrays[0]
                new_args2 = []
                first = True
                for arg in new_args:
                    if arg in boolean_arrays:
                        if first:
                            new_args2.append(new_array)
                            first = False
                    else:
                        new_args2.append(arg)
                new_args = new_args2

            # Mask out integer arrays to only where the start is less than the
            # stop for all arrays.
            if integer_arrays:
                starts, stops = zip(*integer_arrays)
                starts = array(broadcast_arrays(*starts))
                stops = array(broadcast_arrays(*stops))
                mask = logical_and.reduce(starts < stops, axis=0)
                new_args2 = []
                i = 0
                for arg in new_args:
                    if isinstance(arg, IntegerArray):
                        if mask.ndim == 0:
                            # Integer arrays always result in a 1 dimensional
                            # result, except when we have a scalar, we want to
                            # have a 0 dimensional result to match Integer().
                            new_args2.append(IntegerArray(starts[i]))
                        elif mask.all():
                            new_args2.append(IntegerArray(starts[i]))
                        else:
                            new_args2.append(IntegerArray(starts[i, mask]))
                        if new_args2[-1].isempty():
                            raise ValueError("Indices do not intersect")
                        i += 1
                    else:
                        new_args2.append(arg)
                new_args = new_args2
            return Tuple(*new_args)
        raise NotImplementedError(f"Tuple.as_subindex() is not implemented for type '{type(index).__name__}")

    def isempty(self, shape=None):
        if shape is not None:
            return 0 in self.newshape(shape)

        return any(i.isempty() for i in self.args)
